#pragma once

#include <algorithm>
#include <chrono>
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <limits.h>
#include <map>
#include <numeric>
#include <random>
#include <set>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

#ifdef WINDOWS
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "zipf.hpp"

namespace Benchmark {

namespace Device {
struct Cuda;
struct Host;
} // namespace Device

template <typename Device> struct BenchTimer;

namespace Generators {
template <typename Type> std::vector<Type> increasing(int size) {
  std::vector<Type> res(size);
  std::iota(res.begin(), res.end(), 0);
  return res;
}

template <typename Type> std::vector<Type> decreasing(int size) {
  std::vector<Type> res(size);
  std::iota(res.begin(), res.end(), 0);
  std::reverse(res.begin(), res.end());
  return res;
}

template <typename Type> std::vector<Type> shuffle(int size) {
  std::vector<Type> res(size);
  std::iota(res.begin(), res.end(), 0);

  std::random_device rd;
  std::mt19937 g(rd());
  std::shuffle(res.begin(), res.end(), g);

  return res;
}

template <typename Type> std::vector<Type> gaussian(int size) {
  std::vector<Type> res(size);
  std::random_device rd;
  std::mt19937 g(rd());

  std::normal_distribution<> dg(size, size / 2.0);
  for (auto i = 0; i < size; ++i) {
    res[i] = std::abs(dg(g));
  }
  return res;
}

template <typename Type> std::vector<Type> zipf(int size) {
  std::vector<Type> res(size);

  Zipf<Type> zipf((uint64_t)size, 0.98, (uint64_t)time(nullptr));
  for (int i = 0; i < size; i++) {
    res.push_back(zipf.next());
  }

  return res;
}

template <typename Type>
std::vector<Type> almostSorted(int size, int swapCount = 5) {
  std::vector<Type> res(size);
  std::iota(res.begin(), res.end(), 0);

  for (int i = 0; i < swapCount; ++i) {
    size_t targetIndex = rand() % (size - 2);
    std::swap(res[targetIndex], res[targetIndex + 1]);
  }
  return res;
}
} // namespace Generators

#ifdef __NVCC__
template <> struct BenchTimer<Device::Cuda> {
  float mTime;
  cudaEvent_t mStart, mStop;

  std::map<std::string, float> mValues;

  double getTime() { return mTime; }

public:
  void start() {
    cudaEventCreate(&mStart);
    cudaEventCreate(&mStop);
    cudaEventRecord(mStart, nullptr);
  }

  void stop(std::string name) {
    cudaEventRecord(mStop, nullptr);
    cudaEventSynchronize(mStop);
    cudaEventElapsedTime(&mTime, mStart, mStop);

    cudaEventDestroy(mStart);
    cudaEventDestroy(mStop);

    mValues[name] = getTime();
  }

  std::map<std::string, double> getTimers() {
    std::map<std::string, double> res;
    for (auto &it : mValues)
      res[it.first] = (double)it.second;
    return res;
  }
};
#endif

template <> struct BenchTimer<Device::Host> {
  std::chrono::high_resolution_clock::time_point tp;
  std::chrono::high_resolution_clock::duration dur;

  std::map<std::string, double> mValues;

  double getTime() {
    return std::chrono::duration_cast<std::chrono::microseconds>(dur).count() /
           1000.0;
  }

public:
  void start() { tp = std::chrono::high_resolution_clock::now(); }

  void stop(std::string name) {
    dur = std::chrono::high_resolution_clock::now() - tp;
    mValues[name] = getTime();
  }

  std::map<std::string, double> getTimers() { return mValues; }
};

template <typename Device, typename Type, typename Code>
std::map<std::string, double> runner(Code &&code, std::vector<Type> input,
                                     int attempts) {
  std::map<std::string, double> sums;

  if (attempts == 0) {
    return sums;
  }

  for (int attempt = 0; attempt < attempts; ++attempt) {
    Benchmark::BenchTimer<Device> timer;
    code(timer, input);

    for (auto const &it : timer.getTimers()) {
      sums[it.first] += it.second;
    }
  }

  std::map<std::string, double> result;
  for (auto const &it : sums) {
    result[it.first] = it.second / (attempts * 1.);
  }
  return result;
}

std::filesystem::path getExecutablePath() {
  std::string res;
#ifdef WINDOWS
  char result[MAX_PATH];
  res = std::string(result, GetModuleFileName(nullptr, result, MAX_PATH));
#else
  char result[PATH_MAX];
  ssize_t count = readlink("/proc/self/exe", result, PATH_MAX);
  res = std::string(result, (count > 0) ? count : 0);
#endif
  return std::filesystem::path(res).remove_filename();
}

template <typename Device, typename Type, typename Code>
void execRun(const std::string &dataName, Code &&code, int attempts,
             std::map<std::string, std::map<std::string, double>> &row,
             std::vector<Type> data) {
  row[dataName] = runner<Device>(code, data, attempts);
  std::cout << "- " << dataName << "(#" << attempts << ")" << std::endl;
}

template <typename Device, typename Type, typename Code>
void execute(const std::string name, Code &&code, int from = 10, int to = 23,
             int attempts = 1) {

  std::map<std::string, std::stringstream> result;
  std::set<std::string> keys;

  for (int pow = from; pow <= to; ++pow) {
    size_t size = 1 << pow;

    std::map<std::string, std::map<std::string, double>> row;
    std::cout << name << "[" << pow << "]:" << std::endl;

    execRun<Device>("increasing", code, attempts, row,
                    Generators::increasing<Type>(size));

    execRun<Device>("decreasing", code, attempts, row,
                    Generators::decreasing<Type>(size));

    execRun<Device>("shuffle", code, attempts, row,
                    Generators::shuffle<Type>(size));

    execRun<Device>("gaussian", code, attempts, row,
                    Generators::gaussian<Type>(size));

    execRun<Device>("almostSorted", code, attempts, row,
                    Generators::almostSorted<Type>(size));

    bool init = keys.size() == 0;
    if (init) {
      for (auto &col : row) {
        for (auto &pair : col.second) {
          keys.insert(pair.first);
        }
      }
    }

    for (auto &key : keys) {
      std::string table(name + "_" + key);

      if (init) {
        result[table] << "pow";
        for (auto &pair : row) {
          result[table] << "," << pair.first;
        }
        result[table] << std::endl;
      }

      result[table] << pow;
      for (auto &pair : row) {
        result[table] << "," << pair.second[key];
      }
      result[table] << std::endl;
    }
  }

  for (auto const &it : result) {
    std::ofstream file(getExecutablePath() / "../" / "_results" /
                       (it.first + ".csv"));
    file << it.second.str();
  }
}

} // namespace Benchmark
