#pragma once

#include <algorithm>
#include <cassert>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <random>
#include <set>
#include <vector>

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>

template <typename T> class Zipf {
  uint64_t n;

  double mTheta;
  double mAlpha;
  double mThres;

  uint64_t mLastN;

  double mDblN;
  double mZetan;
  double mEta;
  uint64_t mRandState;

  static inline double fastRandD(uint64_t *state) {
    *state = (*state * 0x5deece66dUL + 0xbUL) & ((1UL << 48) - 1);
    return (double)*state / (double)((1UL << 48) - 1);
  }

  // http://martin.ankerl.com/2012/01/25/optimized-approximative-pow-in-c-and-cpp/
  static inline double powApprox(double a, double b) {
    union {
      double d;
      int x[2];
    } u = {a};

    u.x[1] = (int)(b * (u.x[1] - 1072632447) + 1072632447);
    u.x[0] = 0;
    return u.d;
  }

  static inline double zeta(uint64_t lastN, double lastSum, uint64_t n,
                            double theta) {
    if (lastN > n) {
      lastN = 0;
      lastSum = 0.;
    }

    while (lastN < n) {
      lastSum += 1. / powApprox((double)lastN + 1., theta);
      lastN++;
    }
    return lastSum;
  }

public:
  Zipf(uint64_t n, double theta, uint64_t randSeed) {
    assert(n > 0);
    if (theta > 0.992 && theta < 1) {
      fprintf(stderr,
              "theta > 0.992 will be inaccurate due to approximation\n");
    }

    if (theta >= 1. && theta < 40.) {
      fprintf(stderr, "theta in [1., 40.) is not supported\n");
      assert(false);
    }

    assert(theta == -1. || (theta >= 0. && theta < 1.) || theta >= 40.);
    assert(randSeed < (1UL << 48));
    memset(this, 0, sizeof(*this));

    this->n = n;
    this->mTheta = theta;

    if (theta == -1.) {
      randSeed = randSeed % n;
    } else if (theta > 0. && theta < 1.) {
      this->mAlpha = 1. / (1. - theta);
      this->mThres = 1. + powApprox(0.5, theta);
    } else {
      this->mAlpha = 0.;
      this->mThres = 0.;
    }

    this->mLastN = 0;
    this->mZetan = 0.;
    this->mRandState = randSeed;
  }

  T next() {
    if (this->mLastN != this->n) {
      if (this->mTheta > 0. && this->mTheta < 1.) {
        this->mZetan = zeta(this->mLastN, this->mZetan, this->n, this->mTheta);
        this->mEta = (1. - powApprox(2. / (double)this->n, 1. - this->mTheta)) /
                     (1. - zeta(0, 0., 2, this->mTheta) / this->mZetan);
      }
      this->mLastN = this->n;
      this->mDblN = (double)this->n;
    }

    if (this->mTheta == -1.) {
      T v = this->mRandState;
      if (++this->mRandState >= this->n) {
        this->mRandState = 0;
      }
      return v;
    }

    if (this->mTheta == 0.) {
      double u = fastRandD(&this->mRandState);
      return (T)(this->mDblN * u);
    }

    if (this->mTheta >= 40.) {
      return 0;
    }

    // from J. Gray et al. Quickly generating billion-record synthetic
    // databases. In SIGMOD, 1994.
    // double u = erand48(this->rand_state);
    double u = fastRandD(&this->mRandState);
    double uz = u * this->mZetan;
    if (uz < 1.) {
      return 0;
    }

    if (uz < this->mThres) {
      return 1;
    }

    return (T)(this->mDblN *
               powApprox(this->mEta * (u - 1.) + 1., this->mAlpha));
  }
};
