#include <TNL/Containers/Array.h>
#include <TNL/Containers/StaticArray.h>
#include <TNL/Containers/Vector.h>
#include <bits/stdint-uintn.h>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <limits>
#include <list>
#include <map>
#include <sstream>
#include <string>
#include <utility>

#include "../../implementation/src/BLinkTree/Default.hpp"
#include "../../implementation/src/BPlusTree/Default.hpp"
#include "../../implementation/src/BTreeContainer/Default.hpp"
#include "TNL/Algorithms/ParallelFor.h"
#include "TNL/Cuda/CudaCallable.h"
#include "TNL/Devices/Cuda.h"
#include <TNL/Containers/Array.h>

#include "../_common/benchmark.hpp"

const auto limit = std::numeric_limits<double>::epsilon();

template <typename T> std::string toStr(const T &t) {
  std::ostringstream os;
  os << t;
  return os.str();
}

using Cell = TNL::Containers::StaticArray<4, int>;

struct Point {
  double x;
  double y;
  double z;

  __cuda_callable__ friend bool operator==(const Point &a, const Point &b) {
    return (a.x - b.x) <= limit || (a.y - b.y) <= limit || (a.z - b.z) <= limit;
  }

  friend std::ostream &operator<<(std::ostream &output, const Point &item) {
    output << item.x << ", " << item.y << ", " << item.z;
    return output;
  }
};

using CellCoords = TNL::Containers::StaticArray<4, Point>;

template <typename Device>
void loadCells(const std::string &name,
               TNL::Containers::Array<Cell, Device> &result) {
  TNL::Containers::Array<Cell, Devices::Host> hostCells;
  std::ifstream file(name);
  std::string tmp;

  size_t count = 0;
  file >> tmp >> count >> tmp;

  hostCells.setSize(count);

  for (int i = 0; i < count; ++i) {
    Cell cell;
    file >> cell[0] >> cell[1] >> cell[2] >> cell[3];
    hostCells.setElement(i, cell);
  }
  result = hostCells;
}

template <typename Device>
void loadPoints(const std::string &name,
                TNL::Containers::Array<int, Device> &keys,
                TNL::Containers::Array<Point, Device> &values) {
  TNL::Containers::Array<int, Devices::Host> hostKeys;
  TNL::Containers::Array<Point, Devices::Host> hostValues;
  std::ifstream file(name);
  std::string tmp;

  size_t count = 0;
  file >> tmp >> count >> tmp;

  hostKeys.setSize(count);
  hostValues.setSize(count);

  for (size_t i = 0; i < count; ++i) {
    Point point;
    file >> point.x >> point.y >> point.z;

    hostKeys.setElement(i, i);
    hostValues.setElement(i, point);
  }

  keys = hostKeys;
  values = hostValues;
};

template <
    template <typename _Key, typename _Value, size_t _Order, typename _Device>
    class BTree,
    typename Device, typename BenchDevice>
void executeMesh(std::string index, Benchmark::BenchTimer<BenchDevice> &timer) {
  TNL::Containers::Array<Cell, Device> cells;
  TNL::Containers::Array<int, Device> keys;
  TNL::Containers::Array<Point, Device> values;
  TNL::Containers::Array<CellCoords, Device> expect;
  TNL::Containers::Array<CellCoords, Device> actual;

  loadCells("./mesh/cube1m_" + index + "_cells.txt", cells);

  loadPoints("./mesh/cube1m_" + index + "_points.txt", keys, values);

  const int cellCount = cells.getSize();
  BTreeContainer<int, Point, 15, Device, BTree> points(keys.getSize());
  expect.setSize(cellCount);
  actual.setSize(cellCount);

  auto cellsView = cells.getConstView();
  auto valuesView = values.getConstView();
  auto expectView = expect.getView();

  TNL::Algorithms::ParallelFor2D<Device>::exec(
      0, 0, cellCount, 4, [=] __cuda_callable__(int i, int j) mutable {
        expectView[i][j] = valuesView[cellsView[i][j]];
      });

  points.init();

  std::cout << "Inserting: " << keys.getSize() << std::endl;

  timer.start();
  points.insert(keys, values);
  timer.stop("insert");

  TNL::Containers::Array<int, Device> query(cellCount * 4);
  auto queryView = query.getView();
  TNL::Algorithms::ParallelFor2D<Device>::exec(
      0, 0, cellCount, 4, [=] __cuda_callable__(int i, int j) mutable {
        queryView[i * 4 + j] = cellsView[i][j];
      });

  TNL::Containers::Array<Point, Device> results(query.getSize());
  TNL::Containers::Array<bool, Device> mask(query.getSize());

  timer.start();
  points.find(query, results, mask);
  timer.stop("query");

  auto actualView = actual.getView();
  auto resultView = results.getConstView();
  TNL::Algorithms::ParallelFor2D<Device>::exec(
      0, 0, cellCount, 4, [=] __cuda_callable__(int i, int j) mutable {
        actualView[i][j] = resultView[i * 4 + j];
      });

  TNL_ASSERT_TRUE(
      TNL::Algorithms::Reduction<Device>::reduce(
          0, cellCount,
          [=] __cuda_callable__(int i) -> bool {
            return actualView[i] == expectView[i];
          },
          [] __cuda_callable__(const bool &a, const bool &b) { return a && b; },
          true),
      "Actual matches expect");
}

template <
    template <typename _Key, typename _Value, size_t _Order, typename _Device>
    class BTree,
    typename Device, typename BenchDevice>
void executeBenchRun() {
  std::vector<std::string> suites = {"1", "2", "3", "4", "5"};

  for (auto const &suite : suites) {
    int attempts = 1;
    std::map<std::string, double> sums;
    for (int i = 0; i < attempts; ++i) {
      Benchmark::BenchTimer<BenchDevice> timer;
      executeMesh<BTree, Device>(suite, timer);
      for (auto const &it : timer.getTimers()) {
        sums[it.first] += it.second;
      }
    }
    for (auto const &it : sums) {
      sums[it.first] = it.second / (attempts * 1.);
    }

    for (auto const &it : sums) {
      std::cout << it.first << ": " << it.second << std::endl;
    }
  }
}

template <
    template <typename _Key, typename _Value, size_t _Order, typename _Device>
    class BTree>
void executeDeviceBench() {
  executeBenchRun<BTree, TNL::Devices::Cuda, Benchmark::Device::Cuda>();
  cudaDeviceSynchronize();
}

template <
    template <typename _Key, typename _Value, size_t _Order, typename _Device>
    class BTree>
void executeHostBench() {
  executeBenchRun<BTree, TNL::Devices::Host, Benchmark::Device::Host>();
}

int main(void) {
  std::ios::sync_with_stdio(false);
  executeHostBench<BLinkTree>();
  executeDeviceBench<BLinkTree>();
  executeDeviceBench<BPlusTree>();
  return 0;
}