/** @jsxImportSource @emotion/react */
import { useContext, useState } from "react"
import { css } from "@emotion/react"
import { getInstance, MemoryContext } from "../core/node"
import styled from "@emotion/styled"
import { AnimatePresence, motion } from "framer-motion"

const StyleLink = css`
  position: relative;

  &:before {
    content: "";
    top: 0;
    left: 0;
    right: 0;
    bottom: 0;
    position: absolute;
    border: 5px solid red;
  }
`

interface NodeProps {
  address: string
  onAddressClick?: () => void
  onLinkClick?: () => void
  isLink?: boolean
  depth: number
}

const NodeContainer = styled.div`
  display: flex;
  flex-direction: column;
  text-align: left;
  background: rgba(0, 0, 0, 0.1);
  text-shadow: 0px 1px 0px rgba(255, 255, 255, 0.3),
    0px -1px 0px rgba(255, 255, 255, 0.3), 1px 0px 0px rgba(255, 255, 255, 0.3),
    -1px 0px 0px rgba(255, 255, 255, 0.3);

  border: 1px solid rgba(0, 0, 0, 0.3);
  padding: 8px;
  border-radius: 6px;
`

function NodeView(props: NodeProps) {
  const memory = useContext(MemoryContext)
  const inst = getInstance(memory?.memory ?? {}, props.address)

  if (!inst.orphan && props.isLink) return null

  return (
    <NodeContainer css={inst.orphan ? StyleLink : {}}>
      <div
        css={css`
          display: flex;
          justify-content: space-between;
          align-items: center;
        `}
      >
        <strong
          css={css`
            cursor: pointer;
            background-color: ${inst.write ? "yellow" : "none"};
          `}
          onClick={props.onAddressClick}
        >
          {inst.address}
        </strong>
        {inst.sibling && (
          <div
            onClick={!props.isLink ? props.onLinkClick : undefined}
            css={{ fontSize: "0.8em" }}
          >
            ({inst.sibling})
          </div>
        )}
      </div>
      <div css={{ display: "flex", gap: 5 }}>
        <AnimatePresence initial>
          {inst.keys.map((key) => (
            <motion.span
              initial={{ background: "rgba(0, 255, 0, 1)" }}
              animate={{ background: "rgba(0, 255, 0, 0)" }}
              exit={{ background: "rgba(255, 0, 0, 1)" }}
              transition={{ duration: 0.7 }}
              key={key}
            >
              {key}
            </motion.span>
          ))}
        </AnimatePresence>
      </div>
      <div
        css={css`
          display: flex;
          gap: 5px;

          margin-top: 5px;

          &:empty {
            margin-top: 0;
          }
        `}
      >
        {inst.children.map((addr) => {
          return <Node address={addr} key={addr} depth={props.depth + 1} />
        })}
      </div>
    </NodeContainer>
  )
}

export function Node(props: NodeProps) {
  const memory = useContext(MemoryContext)
  const inst = getInstance(memory?.memory ?? {}, props.address)

  const [visible, setVisible] = useState(props.depth < 10)
  const [link, setLink] = useState<boolean>(false)

  return (
    <div css={{ display: "flex" }}>
      {visible ? (
        <>
          <NodeView
            depth={props.depth}
            address={props.address}
            onLinkClick={() => setLink(!link)}
            onAddressClick={() => setVisible(false)}
          />
          {inst.sibling && (
            <NodeView depth={props.depth} address={inst.sibling} isLink />
          )}
        </>
      ) : (
        <NodeContainer>
          <div
            css={{ cursor: "pointer", ...(props.isLink ? StyleLink : {}) }}
            onClick={() => setVisible(true)}
          >
            {props.address}
          </div>
        </NodeContainer>
      )}
    </div>
  )
}
