/** @jsxImportSource @emotion/react */
import { useCallback, useEffect, useState } from "react"
import { css } from "@emotion/react"
import { MemoryContext, Playback, PlaybackState } from "../core/node"
import { PlaybackHistory } from "./PlaybackHistory"
import { RootEditor } from "./RootEditor"
import { Node } from "./NodeView"

export function PlaybackViewer(props: { playback: Playback }) {
  const [userRoot, setUserRoot] = useState("")
  const [state, setState] = useState<PlaybackState>({
    index: -1,
    line: "",
    memory: {},
    validCommand: false,
    root: "",
  })

  const handleLoop = useCallback(
    (count: number) => {
      const actualCount =
        count === Infinity
          ? props.playback.instructions.length -
            Math.max(0, props.playback.index)
          : Math.abs(count)

      let value = null
      for (let i = 0; i < actualCount; ++i) {
        value = props.playback[count > 0 ? "next" : "prev"]()
      }

      if (value) {
        setState(value)
      }
    },
    [props.playback]
  )

  useEffect(() => {
    const listener = (e: KeyboardEvent) => {
      switch (e.key) {
        case "ArrowUp":
          return handleLoop(-1)
        case "ArrowDown":
          return handleLoop(1)
        case "ArrowLeft":
          return handleLoop(-10)
        case "ArrowRight":
          return handleLoop(10)
      }
    }
    window.addEventListener("keydown", listener)
    return () => window.removeEventListener("keydown", listener)
  }, [handleLoop])

  const realRoot = userRoot || state.root

  return (
    <>
      <PlaybackHistory
        playback={props.playback}
        index={state.index}
        onIndexClick={(index) => handleLoop(index - state.index)}
      />
      <div
        css={css`
          display: flex;
          gap: 10px;
          padding: 8px 16px;
        `}
      >
        <pre css={{ flexGrow: 1 }}>
          {state.index + 1} ({state.validCommand ? "valid" : "invalid"}):{" "}
          {state.line}
        </pre>
      </div>

      <div
        css={css`
          display: block;
          width: 100%;
          font-size: 16px;
          overflow-x: auto;
        `}
      >
        <MemoryContext.Provider value={state}>
          <div
            css={css`
              display: inline-flex;
              margin: 0 16px;
            `}
          >
            {realRoot && <Node address={realRoot} depth={0} />}
          </div>
        </MemoryContext.Provider>
      </div>

      <div
        css={css`
          position: absolute;
          bottom: 0;
          left: 0;
          right: 0;
          margin: 16px;
        `}
      >
        <RootEditor value={userRoot} onChange={setUserRoot} />
        <button type="button" onClick={() => handleLoop(-10)}>
          -10
        </button>
        <button type="button" onClick={() => handleLoop(-1)}>
          -1
        </button>
        <button type="button" onClick={() => handleLoop(1)}>
          +1
        </button>
        <button type="button" onClick={() => handleLoop(10)}>
          +10
        </button>
        <button type="button" onClick={() => handleLoop(Infinity)}>
          Play All
        </button>
      </div>
    </>
  )
}
