import { Memory, Instance } from "./types"
import { createContext } from "react"

export function getInstance(memory: Memory, address: string): Instance {
  if (memory[address]) return memory[address]

  return (memory[address] = {
    address,
    write: false,
    keys: [],
    children: [],
    sibling: null,
    orphan: false,
  })
}

function root(memory: Memory, rootAddr: string, order: number) {
  getInstance(memory, rootAddr)
  return rootAddr
}

function remove(memory: Memory, srcAddr: string, index: number) {
  const src = getInstance(memory, srcAddr)
  src.keys.splice(index, 1)
  src.children.splice(index, 1)
}

function increaseHeight(
  memory: Memory,
  srcAddr: string,
  leftAddr: string,
  rightAddr: string,
  key: number
) {
  const src = getInstance(memory, srcAddr)
  const left = getInstance(memory, leftAddr)
  const right = getInstance(memory, rightAddr)

  left.keys = src.keys.slice(0)
  left.children = src.children.slice(0)
  left.sibling = src.sibling

  left.orphan = false
  right.orphan = false

  src.keys = [key]
  src.children = [leftAddr, rightAddr]
  src.sibling = null
}

function split(
  memory: Memory,
  source: string,
  startIdx: number,
  leftLength: number,
  targetAddr: string
) {
  const newItem = getInstance(memory, targetAddr)
  const sourceItem = getInstance(memory, source)

  newItem.keys = sourceItem.keys.slice(startIdx)
  newItem.children = sourceItem.children.slice(startIdx)

  sourceItem.keys = sourceItem.keys.slice(0, leftLength)
  sourceItem.children = getInstance(memory, source).children.slice(
    0,
    leftLength + 1
  )

  newItem.sibling = sourceItem.sibling
  sourceItem.sibling = newItem.address

  newItem.orphan = true
}

function setLatch(memory: Memory, addr: string, write: boolean) {
  getInstance(memory, addr).write = write
}

function insertInternal(
  memory: Memory,
  addr: string,
  key: number,
  childAddr: string,
  index: number,
  childIndex: number
) {
  const target = getInstance(memory, addr)
  const child = getInstance(memory, childAddr)

  child.orphan = false

  target.keys.splice(index, 0, key)
  target.children.splice(childIndex, 0, childAddr)
}

function insertLeaf(memory: Memory, addr: string, key: number, index: number) {
  getInstance(memory, addr).keys.splice(index, 0, key)
}

function transformLog(line: string) {
  function matchAndInvoke<T extends readonly string[]>(
    methods: T,
    callback: (
      method: T[number],
      args: string[]
    ) => ((m: Memory) => void) | null
  ): null | ((m: Memory) => void) {
    const match = line.match(
      new RegExp(`(?<method>${methods.join("|")})\\((?<args>[^\\(\\)]*)\\)`)
    )
    return match?.groups
      ? callback(
          match.groups.method,
          match.groups.args.split(",").map((i) => i.trim())
        )
      : null
  }

  function address(inp: string | undefined | null) {
    if (inp?.startsWith("0x") || inp?.startsWith("Ox")) {
      return inp.substr(2)
    }
    return inp ?? ""
  }

  if (line.includes("LATCH")) {
    return matchAndInvoke(
      ["acquire", "attempt", "release"] as const,
      (method, [addr]) => (memory) =>
        setLatch(memory, address(addr), method !== "release")
    )
  }

  if (line.includes("INSERT")) {
    return matchAndInvoke(
      ["insertInternal", "insertLeaf"] as const,
      (method, args) => {
        if (method === "insertLeaf") {
          if (args.length !== 3) return null

          const [addr, key, index] = args
          return (memory) =>
            insertLeaf(memory, address(addr), Number(key), Number(index))
        }

        if (method === "insertInternal") {
          if (args.length !== 5) return null
          const [addr, key, child, index, childIndex] = args
          return (memory) =>
            insertInternal(
              memory,
              address(addr),
              Number(key),
              address(child),
              Number(index),
              Number(childIndex)
            )
        }

        return null
      }
    )
  }

  if (line.includes("SPLIT")) {
    return matchAndInvoke(
      ["increaseHeight", "split"] as const,
      (method, args) => {
        if (method === "increaseHeight") {
          if (args.length !== 4) return null

          const [addr, left, right, key] = args
          return (memory) =>
            increaseHeight(
              memory,
              address(addr),
              address(left),
              address(right),
              Number(key)
            )
        }

        if (method === "split") {
          if (args.length !== 4) return null
          const [addr, splitIdx, leftLen, target] = args
          return (memory) =>
            split(
              memory,
              address(addr),
              Number(splitIdx),
              Number(leftLen),
              address(target)
            )
        }

        return null
      }
    )
  }

  if (line.includes("DELETE")) {
    return matchAndInvoke(["delete"] as const, (method, args) => {
      if (method === "delete") {
        if (args.length !== 2) return null

        const [addr, index] = args
        return (memory) => remove(memory, address(addr), Number(index))
      }

      return null
    })
  }

  if (line.includes("INIT")) {
    return matchAndInvoke(["root"] as const, (method, args) => {
      if (method === "root") {
        if (args.length !== 2) return null

        const [addr, order] = args
        return (memory) => root(memory, address(addr), Number(order))
      }
      return null
    })
  }

  return null
}

export interface PlaybackState {
  index: number
  line: string
  validCommand: boolean
  memory: Memory
  root: string
}

interface MutableMemoryFunc {
  (memory: Memory): void | string
}

export class Playback {
  index = -1
  private root: string = ""
  private map: Record<number, Memory | null> = { [-1]: {} }
  readonly instructions: [string, MutableMemoryFunc | null][]

  constructor(lines: string[]) {
    this.instructions = lines.map((line) => [line, transformLog(line)])
  }

  private getReelState() {
    const [line, exec] = this.instructions[this.index] ?? []

    return {
      line,
      memory: this.map[this.index] ?? {},
      index: this.index,
      validCommand: exec != null,
      root: this.root,
    }
  }

  prev(): PlaybackState | null {
    if (this.index <= 0) return null
    this.index -= 1

    return this.getReelState()
  }

  next(): PlaybackState | null {
    if (this.index >= this.instructions.length - 1) return null
    this.index += 1

    const [, exec] = this.instructions[this.index] ?? []

    const prevMemory = this.map[this.index - 1]
    if (this.map[this.index] == null && prevMemory != null) {
      let currMemory = prevMemory

      if (exec) {
        currMemory = JSON.parse(JSON.stringify(prevMemory))
        const root = exec(currMemory)

        if (typeof root === "string") {
          this.root = root
        }
      }

      this.map[this.index] = currMemory
    }

    return this.getReelState()
  }
}

export const MemoryContext = createContext<PlaybackState | null>(null)
