#pragma once
#include <TNL/Assert.h>
#include <TNL/Cuda/CudaCallable.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/TypeTraits.h>
#include <iostream>
#include <sstream>
#include <stdint.h>
#include <string>
#include <vector>

#include "../BLinkNode.hpp"

template <typename KeyType, typename ValueType, size_t Order, typename Device>
struct BLinkOperations;

template <typename KeyType, typename ValueType, size_t Order>
struct BLinkOperations<KeyType, ValueType, Order, TNL::Devices::Cuda> {
  using Node = BLinkNode<KeyType, ValueType, Order>;

  template <typename Warp>
  __device__ static void init(Warp &warp, Node *volatile node, bool leaf, Node *volatile sibling,
                              uint8_t highKeyFlag) {

    if (warp.thread_rank() == 0) {
      node->mLeaf = leaf;
      node->mSibling = sibling;
      node->mHighKeyFlag = highKeyFlag;
      node->mSize = 0;
    }
  }

  template <typename Warp>
  __device__ static void init(Warp &warp, Node *volatile node, Node *volatile cursor) {
    BLinkOperations::init(warp, node, cursor->mLeaf, cursor->mSibling, cursor->mHighKeyFlag);
    if (warp.thread_rank() == 0) {
      node->mHighKey = cursor->mHighKey;
    }
  }

  template <typename Warp>
  __device__ static void insertKey(Node *volatile node, size_t index, KeyType key, ValueType value,
                                   Warp &warp) {
    TNL_ASSERT_TRUE(node->mLeaf, "Inserting value is only supported in leaves");

    auto rank = warp.thread_rank();
    if (rank < Order) {
      KeyType thrKey;
      ValueType thrVal;

      if (rank == index) {
        thrKey = key;
        thrVal = value;
      } else if (rank > index) {
        thrKey = node->mKeys[rank - 1];
        thrVal = node->mValues[rank - 1];
      } else {
        thrKey = node->mKeys[rank];
        thrVal = node->mValues[rank];
      }

      node->mKeys[rank] = thrKey;
      node->mValues[rank] = thrVal;

      if (rank == index) {
        node->mSize += 1;
      }
    }
  }

  template <typename Warp>
  __device__ static void insertKey(Node *volatile node, size_t index, KeyType key, Warp &warp) {
    TNL_ASSERT_FALSE(node->mLeaf, "Inserting with no value in an internal node");

    auto rank = warp.thread_rank();
    if (rank < Order) {
      KeyType thrKey;

      if (rank == index) {
        thrKey = key;
      } else if (rank > index) {
        thrKey = node->mKeys[rank - 1];
      } else {
        thrKey = node->mKeys[rank];
      }

      node->mKeys[rank] = thrKey;

      if (rank == index) {
        node->mSize += 1;
      }
    }
  }

  template <typename Warp>
  __device__ static void appendKey(Node *volatile node, KeyType key, ValueType value, Warp &warp) {
    TNL_ASSERT_TRUE(node->mLeaf, "Inserting value is only supported in leaves");

    auto rank = warp.thread_rank();
    if (rank == 0) {
      node->mKeys[node->mSize] = key;
      node->mValues[node->mSize] = value;
      node->mSize += 1;
    }
  }

  template <typename Warp>
  __device__ static void appendKey(Node *volatile node, KeyType key, Warp &warp) {
    TNL_ASSERT_FALSE(node->mLeaf, "Inserting with no value in an internal node");
    auto rank = warp.thread_rank();

    if (rank == 0) {
      node->mKeys[node->mSize] = key;
      node->mSize += 1;
    }
  }

  template <typename Warp>
  __device__ static void removeKey(Node *volatile node, size_t index, Warp &warp) {
    auto rank = warp.thread_rank();
    if (rank < Order) {
      KeyType thrKey;
      ValueType thrVal;

      if (rank >= index) {
        if (rank + 1 < Order) {
          thrKey = node->mKeys[rank + 1];
          thrVal = node->mValues[rank + 1];
        }
      } else {
        thrKey = node->mKeys[rank];
        thrVal = node->mValues[rank];
      }

      if (rank == index) {
        node->mSize -= 1;
      }

      node->mKeys[rank] = thrKey;
      node->mValues[rank] = thrVal;
    }
  }

  template <typename Warp>
  __device__ static void insertChild(Node *volatile node, size_t index, Node *child,
                                     size_t childSize, Warp &warp) {
    auto rank = warp.thread_rank();

    if (rank < Order) {
      Node *thrChild = nullptr;

      if (rank == index) {
        thrChild = child;
      } else if (rank > index) {
        thrChild = node->mChildren[rank - 1];
      } else {
        thrChild = node->mChildren[rank];
      }

      node->mChildren[rank] = thrChild;
    }
  }

  template <typename Warp>
  __device__ static void appendChild(Node *volatile node, Node *volatile child, size_t childSize,
                                     Warp &warp) {
    auto rank = warp.thread_rank();
    if (rank == 0) {
      node->mChildren[childSize] = child;
    }
  }

  template <typename Warp>
  __device__ static void removeChild(Node *volatile node, size_t index, size_t childSize,
                                     Warp &warp) {
    auto rank = warp.thread_rank();
    if (rank < Order) {
      Node *thrChild = nullptr;
      if (rank >= index) {
        if (rank + 1 < Order) {
          thrChild = node->mChildren[rank + 1];
        }
      } else {
        thrChild = node->mChildren[rank];
      }

      node->mChildren[rank] = thrChild;
    }
  }
};