#pragma once
#include <TNL/Cuda/CudaCallable.h>

#include <algorithm>
#include <cstdio>
#include <iostream>
#include <sstream>
#include <string>

#include <cooperative_groups.h>

#include "../Constants.hpp"
#include "../Utils/BumpAllocator.hpp"
#include "../Utils/Debugger.hpp"
#include "../Utils/Print.hpp"

#include "../BNodeLatch/Default.hpp"
#include "./BLinkOperations/Default.hpp"
#include "TNL/Assert.h"

template <typename KeyType, typename ValueType, size_t Order, typename Device> class BLinkTree;

template <typename KeyType, typename ValueType, size_t Order>
class BLinkTree<KeyType, ValueType, Order, Devices::Cuda> {
  using Device = Devices::Cuda;
  using BNode = BLinkNode<KeyType, ValueType, Order>;
  using Latch = BNodeLatch<BNode, Device>;
  using Allocator = BumpAllocator<BNode, Device>;
  using Operations = BLinkOperations<KeyType, ValueType, Order, Device>;

  template <typename Warp>
  __device__ static inline BNode *scan(BNode *node, KeyType key, Warp &warp) {

    if (node->mHighKeyFlag && !(key < node->mHighKey)) {
      return node->mSibling;
    }

    uint16_t size = node->mSize;
    uint16_t childSize = node->childSize();
    uint32_t targetIndex = upperBoundWarp(node->mKeys, size, key, warp);

    if (targetIndex >= childSize) {
      return nullptr;
    }

    BNode *result = node->mChildren[targetIndex];
    return result;
  }

  template <typename Warp>
  __device__ static BNode *splitNode(BNode *cursor, Allocator &alloc, size_t siblingStart,
                                     size_t cursorCount, Warp &warp) {

    auto rank = warp.thread_rank();

    BNode *sibling = (rank == 0) ? alloc.allocate() : nullptr;
    sibling = warp.shfl(sibling, 0);

    Operations::init(warp, sibling, cursor);
    warp.sync();

    uint16_t size = cursor->mSize;
    uint16_t childSize = cursor->childSize();

    if (rank >= siblingStart) {
      if (rank < childSize) {
        sibling->mChildren[rank - siblingStart] = cursor->mChildren[rank];
      }

      if (rank < size) {
        sibling->mKeys[rank - siblingStart] = cursor->mKeys[rank];
        sibling->mValues[rank - siblingStart] = cursor->mValues[rank];
        sibling->mSize = size - siblingStart;
      }
    }

    cursor->mSize = cursorCount;
    cursor->mSibling = sibling;
    cursor->mHighKey = sibling->mKeys[0];
    cursor->mHighKeyFlag = true;

    if (warp.thread_rank() == 0) {
      DBG_SPLIT(cursor, siblingStart, cursorCount, sibling);
    }

    return sibling;
  }

  template <typename Warp>
  __device__ static inline void insertIntoFreeInternal(BNode *curr, KeyType insertKey,
                                                       BNode *insertNode, Warp &warp) {
    auto rank = warp.thread_rank();
    size_t keyIdx = upperBoundWarp(curr->mKeys, curr->mSize, insertKey, warp);
    uint16_t childSize = curr->childSize();

    if (rank == 0) {
      DBG_INSERT_INTERNAL(curr, insertKey, insertNode, keyIdx, (keyIdx + 1));
    }
    Operations::insertChild(curr, keyIdx + 1, insertNode, childSize, warp);
    Operations::insertKey(curr, keyIdx, insertKey, warp);

    warp.sync();
  }

  template <typename Warp>
  __device__ static inline void insertIntoFreeLeaf(BNode *curr, KeyType insertKey, ValueType value,
                                                   BNode *insertNode, Warp &warp) {
    auto rank = warp.thread_rank();
    size_t keyIdx = upperBoundWarp(curr->mKeys, curr->mSize, insertKey, warp);
    uint16_t childSize = curr->childSize();

    if (rank == 0) {
      DBG_INSERT_LEAF(curr, insertKey, keyIdx);
    }
    Operations::insertChild(curr, keyIdx, insertNode, childSize, warp);
    Operations::insertKey(curr, keyIdx, insertKey, value, warp);

    warp.sync();
  }

  template <typename Warp>
  __device__ static inline BNode *increaseTreeHeight(BNode *leftNode, KeyType insertKey,
                                                     BNode *rightNode, Allocator &alloc,
                                                     Warp &warp) {

    BNode *rootNode = nullptr;
    auto rank = warp.thread_rank();
    if (rank == 0) {
      rootNode = alloc.allocate();
    }
    rootNode = warp.shfl(rootNode, 0);

    Operations::init(warp, rootNode, false, nullptr, false);
    warp.sync();

    if (rank == 0) {
      rootNode->mKeys[rank] = insertKey;
      rootNode->mChildren[rank] = leftNode;
    }
    if (rank == 1) {
      rootNode->mChildren[rank] = rightNode;
    }
    rootNode->mSize = 1;

    if (rank == 0) {
      DBG_INCREASE_HEIGHT(leftNode, rootNode, rightNode, insertKey);
    }

    return rootNode;
  }

  template <typename Warp>
  __device__ static inline bool attemptLatchWarp(BNode *node, Latch &latch, Warp &warp) {
    bool attempt = false;

    if (warp.thread_rank() == 0) {
      attempt = latch.attempt(node);
    }

    attempt = warp.any(attempt);
    return attempt;
  }

  template <typename Warp>
  __device__ static inline void releaseLatchWarp(BNode *node, Latch &latch, Warp &warp) {
    if (warp.thread_rank() == 0) {
      latch.release(node);
    }

    warp.sync();
  }

  template <typename Warp>
  __device__ static inline bool verifyParent(BNode *curr, BNode *prev, Warp &warp) {

    auto rank = warp.thread_rank();
    uint16_t childSize = prev->childSize();

    bool found = false;
    if (rank < childSize) {
      found = prev->mChildren[rank] == curr;
    }

    return warp.any(found);
  }

  template <typename Warp>
  __device__ static inline uint32_t lowerBoundWarp(KeyType *haystack, uint32_t haystackSize,
                                                   KeyType needle, Warp &warp) {
    uint64_t rank = warp.thread_rank();
    uint32_t ballot = warp.ballot(!(haystack[rank] < needle));
    uint32_t index = __ffs(ballot) - 1;

    // Enforce one PTX instruction
    asm("min.u32 %0, %1, %2;" : "=r"(index) : "r"(index), "r"(haystackSize));
    return index;
  }

  template <typename Warp>
  __device__ static inline uint32_t upperBoundWarp(KeyType *haystack, uint32_t haystackSize,
                                                   KeyType needle, Warp &warp) {
    uint64_t rank = warp.thread_rank();
    uint32_t ballot = warp.ballot(needle < haystack[rank]);
    uint32_t index = __ffs(ballot) - 1;

    // Enforce one PTX instruction
    asm("min.u32 %0, %1, %2;" : "=r"(index) : "r"(index), "r"(haystackSize));
    return index;
  }

public:
  using NodeType = BNode;
  using OperationsType = Operations;

  template <typename Warp> __device__ static inline BNode *init(Allocator &alloc, Warp &warp) {
    BNode *root = alloc.allocate();
    Operations::init(warp, root, true, nullptr, false);

    DBG_INIT(root, Order);
    return root;
  }

  template <typename Warp>
  __device__ static inline bool find(BNode *root, KeyType key, ValueType &result, Warp &warp) {
    uint64_t rank = warp.thread_rank();

    BNode *next = root;
    while (next) {
      uint16_t isLeaf = next->mLeaf;

      uint32_t size = next->mSize;
      uint32_t ballot = warp.ballot(!(next->mKeys[rank] < key));
      uint32_t targetIdx = min(__ffs(ballot) - 1, size);

      KeyType targetKey = next->mKeys[targetIdx];

      // targetKey == key
      uint32_t foundKey = !(targetKey < key || key < targetKey);

      if (isLeaf) {
        if (foundKey && targetIdx < size) {
          result = next->mValues[targetIdx];
          return true;
        }
        return false;
      }

      // assumption: the tree is fully constructed and insert / splits are done
      next = next->mChildren[targetIdx + foundKey];
    }

    return false;
  }

  template <typename Warp>
  __device__ static bool insert(BNode *&root, KeyType key, ValueType value, Allocator &alloc,
                                Latch &latch, Warp &warp) {
    BNode *prev = nullptr, *curr = nullptr;

    do {
      __threadfence();

      BNode *tmpPrev = prev;
      if (curr == nullptr) {
        curr = root;
      } else {
        prev = curr;
        curr = scan(curr, key, warp);
      }

      if (curr->isFull()) {
        if (attemptLatchWarp(curr, latch, warp) == false) {
          return false;
        }

        uint16_t medianIdx = curr->mSize / 2;
        uint16_t splitIdx = medianIdx + (curr->mLeaf == false);
        KeyType medianKey = curr->mKeys[medianIdx];

        if (curr == root) {
          BNode *split = splitNode(curr, alloc, splitIdx, medianIdx, warp);
          __threadfence();

          root = increaseTreeHeight(curr, medianKey, split, alloc, warp);

          __threadfence();
          releaseLatchWarp(curr, latch, warp);
        } else {
          // latch the parent too
          if (attemptLatchWarp(prev, latch, warp) == false) {
            releaseLatchWarp(curr, latch, warp);
            return false;
          }

          if (prev && (prev->isFull() || verifyParent(curr, prev, warp) == false)) {
            releaseLatchWarp(curr, latch, warp);
            releaseLatchWarp(prev, latch, warp);
            return false;
          }

          BNode *split = splitNode(curr, alloc, splitIdx, medianIdx, warp);

          __threadfence();
          releaseLatchWarp(curr, latch, warp);

          insertIntoFreeInternal(prev, medianKey, split, warp);

          __threadfence();
          releaseLatchWarp(prev, latch, warp);
        }

        curr = prev;
        prev = tmpPrev;
      }
    } while (curr && curr->mLeaf == false);

    if (curr == nullptr || attemptLatchWarp(curr, latch, warp) == false) {
      return false;
    }

    BNode *tmp;
    while (curr->mSibling && (tmp = scan(curr, key, warp)) == curr->mSibling) {
      bool attempt = attemptLatchWarp(tmp, latch, warp);
      releaseLatchWarp(curr, latch, warp);

      if (attempt == false) {
        return false;
      }
      curr = tmp;
    }

    if (curr->isFull()) {
      releaseLatchWarp(curr, latch, warp);
      return false;
    }

    insertIntoFreeLeaf(curr, key, value, nullptr, warp);

    __threadfence();
    releaseLatchWarp(curr, latch, warp);
    return true;
  }

  template <typename Warp>
  __device__ static inline bool remove(BNode *root, KeyType key, Latch &latch, Warp &warp) {
    BNode *tmp;
    BNode *cursor = root;

    while (cursor && cursor->mLeaf == false) {
      cursor = scan(cursor, key, warp);
    }

    while (cursor && cursor->mSibling && (tmp = scan(cursor, key, warp)) == cursor->mSibling) {
      cursor = tmp;
    }

    if (!cursor) {
      return true;
    }

    if (attemptLatchWarp(cursor, latch, warp) == false) {
      return false;
    }

    __threadfence();

    uint16_t size = cursor->mSize;
    uint16_t keyIdx = lowerBoundWarp(cursor->mKeys, size, key, warp);

    if (!(keyIdx < cursor->mSize)) {
      releaseLatchWarp(cursor, latch, warp);
      return true;
    }

    KeyType cursorKey = cursor->mKeys[keyIdx];
    // cursorKey != key
    if (cursorKey < key || key < cursorKey) {
      releaseLatchWarp(cursor, latch, warp);
      return true;
    }

    Operations::removeChild(cursor, keyIdx, cursor->childSize(), warp);
    Operations::removeKey(cursor, keyIdx, warp);
    warp.sync();

    if (warp.thread_rank() == 0) {
      DBG_DELETE(cursor, keyIdx);
    }

    releaseLatchWarp(cursor, latch, warp);
    return true;
  }
};
