#pragma once
#define BTREE_HOST 1

#include <TNL/Cuda/CudaCallable.h>

#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>

#include "../Constants.hpp"
#include "../Utils/Algorithms.hpp"
#include "../Utils/BumpAllocator.hpp"
#include "../Utils/Print.hpp"

#include "../BNodeLatch/Default.hpp"
#include "./BLinkNode.hpp"
#include "./BLinkOperations/Default.hpp"

template <typename KeyType, typename ValueType, size_t Order, typename Device> class BLinkTree;

template <typename KeyType, typename ValueType, size_t Order>
class BLinkTree<KeyType, ValueType, Order, Devices::Host> {
  using Device = Devices::Host;
  using Node = BLinkNode<KeyType, ValueType, Order>;
  using Latch = BNodeLatch<Node, Device>;
  using Allocator = BumpAllocator<Node, Device>;
  using Operations = BLinkOperations<KeyType, ValueType, Order, Device>;

  static inline Node *scan(Node *node, KeyType key) {
    if (node->mHighKeyFlag && !(key < node->mHighKey)) {
      return node->mSibling;
    }

    size_t size = node->mSize;
    size_t childSize = node->childSize();
    size_t targetIndex = upperBound(node->mKeys, size, key);

    if (targetIndex == node->childSize()) {
      return nullptr;
    }

    Node *result = node->mChildren[targetIndex];
    return result;
  }

  static inline Node *moveSide(Node *cursor, KeyType key) {
    Node *tmp, *result = cursor;
    while (result != nullptr && result->mSibling != nullptr &&
           (tmp = scan(result, key)) == result->mSibling) {
      result = tmp;
    }

    return result;
  }

  static inline Node *findLeaf(Node *root, KeyType key) {
    Node *cursor = root;
    while (cursor != nullptr && cursor->mLeaf == false) {
      cursor = scan(cursor, key);
    }

    cursor = moveSide(cursor, key);
    return cursor;
  }

  static Node *splitNode(Node *cursor, Allocator &alloc, size_t siblingStart, size_t cursorCount) {

    Node *sibling = alloc.allocate();
    Operations::init(sibling, cursor);

    for (size_t i = siblingStart; i < cursor->childSize(); ++i) {
      Node *node = cursor->mChildren[i];
      Operations::appendChild(sibling, node, i - siblingStart);
    }

    for (size_t i = siblingStart; i < cursor->mSize; ++i) {
      if (sibling->mLeaf) {
        Operations::appendKey(sibling, cursor->mKeys[i], cursor->mValues[i]);
      } else {
        Operations::appendKey(sibling, cursor->mKeys[i]);
      }
    }

    cursor->mSize = cursorCount;
    cursor->mSibling = sibling;
    cursor->mHighKey = sibling->mKeys[0];
    cursor->mHighKeyFlag = true;

    return sibling;
  }

  static inline void insertIntoFreeInternal(Node *curr, KeyType insertKey, Node *insertNode) {
    size_t keyIdx = upperBound(curr->mKeys, curr->mSize, insertKey);
    Operations::insertChild(curr, keyIdx + 1, insertNode, curr->childSize());
    Operations::insertKey(curr, keyIdx, insertKey);

    if (curr->mLeaf) {

    } else {
    }
  }

  static inline void insertIntoFreeLeaf(Node *curr, KeyType insertKey, ValueType value,
                                        Node *insertNode) {
    size_t keyIdx = upperBound(curr->mKeys, curr->mSize, insertKey);
    Operations::insertChild(curr, keyIdx, insertNode, curr->childSize());
    Operations::insertKey(curr, keyIdx, insertKey, value);
  }

  static inline void increaseTreeHeight(Node *curr, KeyType insertKey, Node *insertNode,
                                        Allocator &alloc) {

    Node *leftNode = alloc.allocate();

    // copy the entire node to the element
    Operations::init(leftNode, curr);
    for (size_t childKey = 0; childKey < curr->childSize(); ++childKey) {
      Node *child = curr->mChildren[childKey];
      Operations::appendChild(leftNode, child, childKey);

      if (childKey < curr->mSize) {
        if (leftNode->mLeaf) {
          Operations::appendKey(leftNode, curr->mKeys[childKey], curr->mValues[childKey]);
        } else {
          Operations::appendKey(leftNode, curr->mKeys[childKey]);
        }
      }
    }

    // this might cause some issues in synchronization
    Operations::init(curr, false, nullptr, false);
    Operations::appendChild(curr, leftNode, 0);
    Operations::appendKey(curr, insertKey);
    Operations::appendChild(curr, insertNode, 1);
  }

public:
  using NodeType = Node;
  using OperationsType = Operations;

  static inline Node *init(Allocator &alloc) {
    Node *root = alloc.allocate();
    Operations::init(root, true, nullptr, false);
    return root;
  }

  static inline bool find(Node *root, KeyType key, ValueType &result) {
    Node *leaf = findLeaf(root, key);
    if (leaf == nullptr)
      return false;

    size_t it = lowerBound(leaf->mKeys, leaf->mSize, key);
    if (it >= leaf->mSize) {
      return false;
    }

    KeyType leafKey = leaf->mKeys[it];
    // leafKey == key
    if (!(leafKey < key || key < leafKey)) {
      result = leaf->mValues[it];
      return true;
    }
    return false;
  }

  static bool insert(Node *root, KeyType key, ValueType value, Allocator &alloc, Latch &latch) {
    Node *prev = nullptr, *curr = nullptr;
    do {
      Node *tmpPrev = prev;
      if (curr == nullptr) {
        curr = root;
      } else {
        prev = curr;
        curr = scan(curr, key);
      }

      if (curr != nullptr && curr->mSize + !curr->mLeaf >= Order) {
        // if node is full, we haven't propagated, restart the operation
        if (curr->mSibling == prev && curr != root) {
          return false;
        }

        size_t medianIdx = curr->mSize / 2;
        size_t splitIdx = medianIdx + (!curr->mLeaf);

        KeyType medianKey = curr->mKeys[medianIdx];
        Node *split = nullptr;

        split = splitNode(curr, alloc, splitIdx, medianIdx);

        if (curr == root) {
          increaseTreeHeight(curr, medianKey, split, alloc);
        } else {
          insertIntoFreeInternal(prev, medianKey, split);
          curr = prev;
          prev = tmpPrev;
        }
      }
    } while (curr != nullptr && curr->mLeaf == false);

    if (curr == nullptr) {
      return false;
    }

    Node *tmp;
    while (curr->mSibling != nullptr && (tmp = scan(curr, key)) == curr->mSibling) {
      curr = tmp;
    }

    insertIntoFreeLeaf(curr, key, value, nullptr);
    return true;
  }

  static inline bool remove(Node *root, KeyType key, Latch &latch) {
    Node *cursor = findLeaf(root, key);

    if (cursor == nullptr)
      return true;

    if (!latch.attempt(cursor)) {
      return false;
    }

    // remove item from keys and remove the respective node
    size_t keyIdx = lowerBound(cursor->mKeys, cursor->mSize, key);

    if (keyIdx == cursor->mSize) {
      latch.release(cursor);
      return true;
    }

    KeyType cursorKey = cursor->mKeys[keyIdx];

    // cursorKey != key
    if (cursorKey < key || key < cursorKey) {
      latch.release(cursor);
      return true;
    }

    Operations::removeChild(cursor, keyIdx, cursor->childSize());
    Operations::removeKey(cursor, keyIdx);

    latch.release(cursor);
    return true;
  }
};