#pragma once
#include <TNL/Cuda/CudaCallable.h>

#include <algorithm>
#include <bits/stdint-uintn.h>
#include <iostream>
#include <sstream>
#include <string>

#include <cooperative_groups.h>

#include "../Constants.hpp"
#include "../Utils/BumpAllocator.hpp"
#include "../Utils/Debugger.hpp"
#include "../Utils/Print.hpp"

#include "../BNodeLatch/Default.hpp"

#include "./BPlusOperations/Default.hpp"
#include "./BPlusNode.hpp"
#include "TNL/Assert.h"

template <typename KeyType, typename ValueType, size_t Order, typename Device> class BPlusTree;

template <typename KeyType, typename ValueType, size_t Order>
class BPlusTree<KeyType, ValueType, Order, Devices::Cuda> {
  using Device = Devices::Cuda;
  using Node = BPlusNode<KeyType, ValueType, Order>;
  using Latch = BNodeLatch<Node, Device>;
  using Allocator = BumpAllocator<Node, Device>;
  using Operations = BPlusOperations<KeyType, ValueType, Order, Device>;

public:
  using NodeType = Node;
  using OperationsType = Operations;

  template <typename Warp>
  __device__ static inline Node *scan(Node *node, KeyType key, Warp &warp) {

    if (node->mHighKeyFlag && !(key < node->mHighKey)) {
      return node->mSibling;
    }

    uint16_t size = node->mSize;
    uint16_t childSize = node->childSize();
    uint32_t targetIndex = upperBoundWarp(node->mKeys, size, key, warp);

    if (targetIndex >= childSize) {
      return nullptr;
    }

    Node *result = node->mChildren[targetIndex];
    return result;
  }

  template <typename Warp>
  __device__ static Node *splitNode(Node *cursor, Allocator &alloc, size_t siblingStart,
                                    size_t cursorCount, Warp &warp) {

    auto rank = warp.thread_rank();

    Node *sibling = (rank == 0) ? alloc.allocate() : nullptr;
    sibling = warp.shfl(sibling, 0);

    Operations::init(warp, sibling, cursor);

    uint16_t size = cursor->mSize;
    uint16_t childSize = cursor->childSize();

    if (rank >= siblingStart) {
      if (rank < childSize) {
        sibling->mChildren[rank - siblingStart] = cursor->mChildren[rank];
      }

      if (rank < size) {
        sibling->mKeys[rank - siblingStart] = cursor->mKeys[rank];
        sibling->mValues[rank - siblingStart] = cursor->mValues[rank];
        sibling->mSize = size - siblingStart;
      }
    }

    cursor->mSize = cursorCount;
    cursor->mSibling = sibling;
    cursor->mHighKey = sibling->mKeys[0];
    cursor->mHighKeyFlag = true;

    if (warp.thread_rank() == 0) {
      DBG_SPLIT(cursor, siblingStart, cursorCount, sibling);
    }

    return sibling;
  }

  template <typename Warp>
  __device__ static inline void insertIntoFreeInternal(Node *curr, KeyType insertKey,
                                                       Node *insertNode, Warp &warp) {
    auto rank = warp.thread_rank();
    size_t keyIdx = upperBoundWarp(curr->mKeys, curr->mSize, insertKey, warp);

    if (rank == 0) {
      DBG_INSERT_INTERNAL(curr, insertKey, insertNode, keyIdx, (keyIdx + 1));
    }
    Operations::insertChild(curr, keyIdx + 1, insertNode, curr->childSize(), warp);
    Operations::insertKey(curr, keyIdx, insertKey, warp);
  }

  template <typename Warp>
  __device__ static inline void insertIntoFreeLeaf(Node *curr, KeyType insertKey, ValueType value,
                                                   Node *insertNode, Warp &warp) {
    auto rank = warp.thread_rank();
    size_t keyIdx = upperBoundWarp(curr->mKeys, curr->mSize, insertKey, warp);

    if (rank == 0) {
      DBG_INSERT_LEAF(curr, insertKey, keyIdx);
    }
    Operations::insertChild(curr, keyIdx, insertNode, curr->childSize(), warp);
    Operations::insertKey(curr, keyIdx, insertKey, value, warp);
  }

  template <typename Warp>
  __device__ static inline void increaseTreeHeight(Node *curr, KeyType insertKey, Node *insertNode,
                                                   Allocator &alloc, Warp &warp) {

    Node *leftNode = nullptr;
    auto rank = warp.thread_rank();
    if (rank == 0) {
      leftNode = alloc.allocate();
    }
    leftNode = warp.shfl(leftNode, 0);

    // copy children back
    uint16_t size = curr->mSize;
    uint16_t childSize = curr->childSize();

    // copy the root to left
    Operations::init(warp, leftNode, curr);
    if (rank < childSize) {
      leftNode->mChildren[rank] = curr->mChildren[rank];
    }
    if (rank < size) {
      leftNode->mKeys[rank] = curr->mKeys[rank];
      leftNode->mValues[rank] = curr->mValues[rank];
    }
    leftNode->mSize = size;

    // update the root
    Operations::init(warp, curr, false, nullptr, false);
    if (rank == 0) {
      curr->mKeys[rank] = insertKey;
      curr->mChildren[rank] = leftNode;
    }
    if (rank == 1) {
      curr->mChildren[rank] = insertNode;
    }
    curr->mSize = 1;

    if (rank == 0) {
      DBG_INCREASE_HEIGHT(curr, leftNode, insertNode, insertKey);
    }
  }

  template <typename Warp>
  __device__ static inline bool attemptLatchWarp(Node *node, Latch &latch, Warp &warp) {
    bool attempt = false;

    if (warp.thread_rank() == 0) {
      attempt = latch.attempt(node);
    }

    attempt = warp.any(attempt);
    return attempt;
  }

  template <typename Warp>
  __device__ static inline void releaseLatchWarp(Node *node, Latch &latch, Warp &warp) {
    if (warp.thread_rank() == 0) {
      latch.release(node);
    }

    warp.sync();
  }

  template <typename Warp>
  __device__ static inline bool verifyParent(Node *curr, Node *prev, Warp &warp) {

    auto rank = warp.thread_rank();
    uint16_t childSize = prev->childSize();

    bool found = false;
    if (rank < childSize) {
      found = prev->mChildren[rank] == curr;
    }

    return warp.any(found);
  }

  template <typename Warp>
  __device__ static inline uint32_t lowerBoundWarp(KeyType *haystack, uint32_t haystackSize,
                                                   KeyType needle, Warp &warp) {
    uint64_t rank = warp.thread_rank();
    uint32_t ballot = warp.ballot(!(haystack[rank] < needle));
    uint32_t index = __ffs(ballot) - 1;

    // Enforce one PTX instruction
    asm("min.u32 %0, %1, %2;" : "=r"(index) : "r"(index), "r"(haystackSize));
    return index;
  }

  template <typename Warp>
  __device__ static inline uint32_t upperBoundWarp(KeyType *haystack, uint32_t haystackSize,
                                                   KeyType needle, Warp &warp) {
    uint64_t rank = warp.thread_rank();
    uint32_t ballot = warp.ballot(needle < haystack[rank]);
    uint32_t index = __ffs(ballot) - 1;

    // Enforce one PTX instruction
    asm("min.u32 %0, %1, %2;" : "=r"(index) : "r"(index), "r"(haystackSize));
    return index;
  }

  template <typename Warp> __device__ static inline Node *init(Allocator &alloc, Warp &warp) {
    Node *root = alloc.allocate();
    Operations::init(warp, root, true, nullptr, false);

    DBG_INIT(root, Order);
    return root;
  }

  template <typename Warp>
  __device__ static inline bool find(Node *root, KeyType key, ValueType &result, Warp &warp) {
    uint64_t rank = warp.thread_rank();

    Node *next = root;
    while (next) {
      uint16_t isLeaf = next->mLeaf;
      uint32_t size = next->mSize;

      // next->mKeys[rank] >= key
      uint32_t ballot = warp.ballot(!(next->mKeys[rank] < key));
      uint32_t targetIdx = __ffs(ballot) - 1;
      asm("min.u32 %0, %1, %2;" : "=r"(targetIdx) : "r"(targetIdx), "r"(size));

      KeyType targetKey = next->mKeys[targetIdx];

      // targetKey == key
      uint32_t foundKey = !(targetKey < key || key < targetKey);

      if (isLeaf) {
        if (foundKey && targetIdx < size) {
          result = next->mValues[targetIdx];
          return true;
        }
        return false;
      }

      // assumption: the tree is fully constructed and insert / splits are done
      next = next->mChildren[targetIdx + foundKey];
    }

    return false;
  }

  template <typename Warp>
  __device__ static bool insert(Node *root, KeyType key, ValueType value, Allocator &alloc,
                                Latch &latch, Warp &warp) {
    Node *prev = nullptr, *curr = nullptr;

    do {
      __threadfence();

      Node *tmpPrev = prev;
      if (curr == nullptr) {
        curr = root;
      } else {
        prev = curr;
        curr = scan(curr, key, warp);
      }

      // restart the operation if we found a lock node
      if (curr->mWriteLock) {
        return false;
      }

      if (curr->isFull()) {
        if (attemptLatchWarp(curr, latch, warp) == false) {
          return false;
        }

        __threadfence();

        size_t medianIdx = curr->mSize / 2;
        size_t splitIdx = medianIdx + (!curr->mLeaf);
        KeyType medianKey = curr->mKeys[medianIdx];

        if (curr == root) {
          Node *split = splitNode(curr, alloc, splitIdx, medianIdx, warp);
          increaseTreeHeight(curr, medianKey, split, alloc, warp);

          __threadfence();
          releaseLatchWarp(curr, latch, warp);
        } else {
          // latch the parent too
          if (attemptLatchWarp(prev, latch, warp) == false) {
            releaseLatchWarp(curr, latch, warp);
            return false;
          }

          __threadfence();

          // verify, if the parent is actually our parent
          if (verifyParent(curr, prev, warp) == false) {
            releaseLatchWarp(curr, latch, warp);
            releaseLatchWarp(prev, latch, warp);
            return false;
          }

          Node *split = splitNode(curr, alloc, splitIdx, medianIdx, warp);

          __threadfence();
          releaseLatchWarp(curr, latch, warp);

          insertIntoFreeInternal(prev, medianKey, split, warp);

          __threadfence();
          releaseLatchWarp(prev, latch, warp);
        }

        // restart the search from root
        curr = prev;
        prev = tmpPrev;
      }
    } while (curr && curr->mLeaf == false);

    if (curr == nullptr) {
      return false;
    }

    if (attemptLatchWarp(curr, latch, warp) == false) {
      return false;
    }

    __threadfence();

    // here we don't check if the node has subsequently become full
    if (curr->isFull()) {
      releaseLatchWarp(curr, latch, warp);
      return false;
    }

    insertIntoFreeLeaf(curr, key, value, nullptr, warp);

    __threadfence();
    releaseLatchWarp(curr, latch, warp);
    return true;
  }

  template <typename Warp>
  __device__ static inline bool remove(Node *root, KeyType key, Latch &latch, Warp &warp) {
    Node *tmp;
    Node *cursor = root;

    while (cursor && cursor->mLeaf == false) {
      cursor = scan(cursor, key, warp);
    }

    while (cursor && cursor->mSibling && (tmp = scan(cursor, key, warp)) == cursor->mSibling) {
      cursor = tmp;
    }

    if (!cursor) {
      return true;
    }

    // remove item from keys and remove the respective node
    if (attemptLatchWarp(cursor, latch, warp) == false) {
      return false;
    }

    __threadfence();

    uint16_t size = cursor->mSize;
    uint16_t keyIdx = lowerBoundWarp(cursor->mKeys, size, key, warp);

    if (!(keyIdx < cursor->mSize)) {
      releaseLatchWarp(cursor, latch, warp);
      return true;
    }

    KeyType cursorKey = cursor->mKeys[keyIdx];
    // cursorKey != key
    if (cursorKey < key || key < cursorKey) {
      releaseLatchWarp(cursor, latch, warp);
      return true;
    }

    Operations::removeChild(cursor, keyIdx, size, warp);
    Operations::removeKey(cursor, keyIdx, warp);

    if (warp.thread_rank() == 0) {
      DBG_DELETE(cursor, keyIdx);
    }

    __threadfence();

    releaseLatchWarp(cursor, latch, warp);
    return true;
  }
};
