#pragma once
#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Assert.h>
#include <TNL/Containers/Array.h>
#include <TNL/Cuda/CudaCallable.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/Devices/Host.h>
#include <TNL/Exceptions/CudaSupportMissing.h>
#include <TNL/Exceptions/NotImplementedError.h>

#include <cooperative_groups.h>
#include <type_traits>

#include "../BNodeLatch/Default.hpp"
#include "../Utils/Algorithms.hpp"
#include "../Utils/BumpAllocator.hpp"
#include "../Utils/DeviceVariable.hpp"

namespace cg = cooperative_groups;

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeInitKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        root,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        alloc) {
  using Tree = BTree<Key, Value, Order, Devices::Cuda>;
  using Node = typename Tree::NodeType;

  auto warp = cg::tiled_partition<32>(cg::this_thread_block());
  root = Tree::init(alloc, warp);
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__device__ static void BTreeInsertWarp(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        &rootPtr,
    bool &toInsert, Key key, Value value,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        &allocator,
    BNodeLatch<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda> &latch) {

  using Tree = BTree<Key, Value, Order, Devices::Cuda>;
  using Node = typename Tree::NodeType;

  auto warp = cg::tiled_partition<32>(cg::this_thread_block());
  TNL_ASSERT_GE(warp.size(), Order, "Found a warp smaller than order");

  uint queue;
  while ((queue = warp.ballot(toInsert))) {
    auto currentLane = __ffs(queue) - 1;
    auto rank = warp.thread_rank();

    Key currentKey = warp.shfl(key, currentLane);
    Value currentValue = warp.shfl(value, currentLane);

    Node *oldRoot = rootPtr;
    Node *root = oldRoot;
    bool insertStatus = Tree::insert(root, currentKey, currentValue, allocator, latch, warp);
    insertStatus = warp.any(insertStatus);

    if (oldRoot != root) {
      rootPtr = root;
    }

    if (warp.thread_rank() == currentLane) {
      toInsert = insertStatus == false;
    }
  }
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeInsertKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        root,
    Containers::ArrayView<std::add_const_t<Key>, Devices::Cuda> keys,
    Containers::ArrayView<Value, Devices::Cuda> values,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        allocator,
    BNodeLatch<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda> latch) {
  auto threadId = threadIdx.x + blockIdx.x * blockDim.x;

  bool toInsert = false;
  Key key;
  Value value;

  if (threadId < keys.getSize()) {
    toInsert = true;
    key = keys.getElement(threadId);
    value = values.getElement(threadId);
  }

  BTreeInsertWarp<Key, Value, Order, BTree>(root, toInsert, key, value, allocator, latch);
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__device__ static void
BTreeFindWarp(DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *,
                             Devices::Cuda> &rootPtr,
              bool &toFind, Key key, Containers::ArrayView<Value, Devices::Cuda> &result,
              Containers::ArrayView<bool, Devices::Cuda> &mask, size_t resultIndex) {

  using Tree = BTree<Key, Value, Order, Devices::Cuda>;
  using Node = typename Tree::NodeType;

  auto warp = cg::tiled_partition<32>(cg::this_thread_block());
  uint queue = 0;
  while ((queue = warp.ballot(toFind))) {
    auto currentLane = __ffs(queue) - 1;
    auto rank = warp.thread_rank();

    Key currentKey = warp.shfl(key, currentLane);
    Value currentRes;
    bool found = Tree::find(rootPtr, currentKey, currentRes, warp);

    if (rank == currentLane) {
      mask[resultIndex] = found;
      if (found) {
        result[resultIndex] = currentRes;
      }

      toFind = false;
    }
  }
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeFindKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        root,
    Containers::ArrayView<std::add_const_t<Key>, Devices::Cuda> keys,
    Containers::ArrayView<Value, Devices::Cuda> result,
    Containers::ArrayView<bool, Devices::Cuda> mask) {

  auto threadId = threadIdx.x + blockIdx.x * blockDim.x;

  bool toFind = false;
  Key key;

  size_t keySize = keys.getSize();
  if (threadId < keySize) {
    toFind = true;
    key = keys.getElement(threadId);
  }

  BTreeFindWarp<Key, Value, Order, BTree>(root, toFind, key, result, mask, threadId);
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__device__ static void BTreeRemoveWarp(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        &rootPtr,
    bool &toRemove, Key key,
    BNodeLatch<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda> latch) {
  using Tree = BTree<Key, Value, Order, Devices::Cuda>;

  auto warp = cg::tiled_partition<32>(cg::this_thread_block());

  uint queue;
  while ((queue = warp.ballot(toRemove))) {
    auto currentLane = __ffs(queue) - 1;

    auto rank = warp.thread_rank();
    Key currentKey = warp.shfl(key, currentLane);

    bool removeStatus = Tree::remove(*rootPtr, currentKey, latch, warp);

    removeStatus = warp.any(removeStatus);
    if (warp.thread_rank() == currentLane) {
      toRemove = removeStatus == false;
    }
  }
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeRemoveKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        rootPtr,
    Containers::ArrayView<std::add_const_t<Key>, Devices::Cuda> keys,
    BNodeLatch<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda> latch) {
  auto threadId = threadIdx.x + blockIdx.x * blockDim.x;

  bool toRemove = false;
  Key key;

  if (threadId < keys.getSize()) {
    toRemove = true;
    key = keys.getElement(threadId);
  }

  BTreeRemoveWarp<Key, Value, Order, BTree>(rootPtr, toRemove, key, latch);
};

template <typename Key, typename Value, size_t Offset, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeBulkCreateInternalKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        baseRoot,
    Containers::ArrayView<std::add_const_t<Key>, Devices::Cuda> inputKeys,
    Containers::ArrayView<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *,
                          Devices::Cuda>
        inputChildren,
    Containers::ArrayView<Key, Devices::Cuda> resultKeys,
    Containers::ArrayView<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *,
                          Devices::Cuda>
        resultChildren,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        allocator) {

  using Tree = BTree<Key, Value, Order, Devices::Cuda>;
  using Node = typename Tree::Node;

  const size_t NodeOffsetSize = Order - Offset;
  const size_t NodeLastSize = NodeOffsetSize / 2;

  size_t idx = threadIdx.x + blockIdx.x * blockDim.x;
  size_t keySize = inputKeys.getSize();
  size_t childSize = inputChildren.getSize();

  if (idx < childSize) {
    Node *child = inputChildren.getElement(idx);
    Key key = (idx < keySize) ? inputKeys.getElement(idx) : 0;

    Key highKey;
    bool highKeyFlag = false;

    size_t idxNode = idx / NodeOffsetSize;

    size_t idxBegin = idxNode * NodeOffsetSize;
    size_t idxEnd = (idxNode + 1) * NodeOffsetSize;

    if (idxEnd == keySize) {
      if ((idx - idxBegin) > NodeLastSize) {
        idxNode += 1;
        idxBegin = idxEnd - NodeLastSize;
      } else {
        idxEnd -= NodeLastSize;
      }
    } else if (idxBegin == keySize) {
      idxEnd = idxBegin;
      idxBegin -= NodeLastSize;
    }

    if (idxEnd > keySize)
      idxEnd = keySize;

    size_t nodeKeySize = (idxEnd - idxBegin) - (idxEnd < keySize);

    Node *targetNode = baseRoot + idxNode;
    Node *siblingNode = nullptr;

    if (idxEnd < keySize) {
      highKey = inputKeys.getElement(idxEnd);
      highKeyFlag = true;
      siblingNode = targetNode + 1;
    }

    size_t idxInsert = idx - idxBegin;
    if (idxInsert == 0) {
      Key lowKey = (idxBegin > 0) ? inputKeys.getElement(idxBegin - 1) : 0;

      targetNode->mSibling = siblingNode;
      targetNode->mHighKey = highKey;
      targetNode->mHighKeyFlag = highKeyFlag;
      targetNode->mLeaf = 0;
      targetNode->mSize = nodeKeySize;

      resultKeys.setElement(idxNode, lowKey);
    }

    targetNode->mKeys[idxInsert] = key;
    targetNode->mChildren[idxInsert] = child;

    resultChildren.setElement(idxNode, targetNode);
  }
};

template <typename Key, typename Value, size_t Offset, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeBulkCreateLeafKernel(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        baseRoot,
    Containers::ArrayView<std::add_const_t<Key>, Devices::Cuda> inputKeys,
    Containers::ArrayView<Value, Devices::Cuda> inputValues,
    Containers::ArrayView<Key, Devices::Cuda> resultKeys,
    Containers::ArrayView<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *,
                          Devices::Cuda>
        resultChildren,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        allocator) {

  using BTreeNode = typename BTree<Key, Value, Order, Devices::Cuda>::NodeType;
  const size_t RealFull = Order - Offset;

  size_t idx = threadIdx.x + blockIdx.x * blockDim.x;
  size_t keySize = inputKeys.getSize();
  size_t valueSize = inputValues.getSize();

  if (idx < keySize) {
    size_t idxNode = idx / RealFull;
    size_t idxInsert = idx - (idxNode * RealFull);

    size_t idxStart = idxNode * RealFull;
    size_t idxSibling = (idxNode + 1) * RealFull;

    BTreeNode *targetNode = baseRoot + idxNode;
    BTreeNode *siblingNode = nullptr;

    Key key = inputKeys.getElement(idx);
    Key highKey;
    bool highKeyFlag = false;

    if (idxSibling < keySize) {
      highKey = inputKeys.getElement(idxSibling);
      highKeyFlag = true;
      siblingNode = targetNode + 1;
    } else {
      idxSibling = keySize;
    }

    targetNode->mKeys[idxInsert] = key;
    targetNode->mValues[idxInsert] = inputValues.getElement(idx);

    if (idxInsert == 0) {
      targetNode->mSibling = siblingNode;
      targetNode->mHighKey = highKey;
      targetNode->mHighKeyFlag = highKeyFlag;
      targetNode->mLeaf = 1;
      targetNode->mSize = idxSibling - idxStart;

      resultKeys.setElement(idxNode, key);
    }

    resultChildren.setElement(idxNode, targetNode);
  }
};

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename Device> class BTree>
__global__ static void BTreeBulkAllocate(
    DeviceVariable<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType *, Devices::Cuda>
        baseRoot,
    size_t size,
    BumpAllocator<typename BTree<Key, Value, Order, Devices::Cuda>::NodeType, Devices::Cuda>
        allocator) {
  baseRoot = allocator.allocate(size);
}

template <typename Key, typename Value, size_t Order, typename Device,
          template <typename _Key, typename _Value, size_t _Order, typename _Device> class BTree>
class BTreeContainer;

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename _Device> class BTree>
class BTreeContainer<Key, Value, Order, Devices::Cuda, BTree> {

  using Device = Devices::Cuda;
  using Tree = BTree<Key, Value, Order, Device>;
  using Node = typename Tree::NodeType;
  using Latch = BNodeLatch<Node, Device>;
  using Allocator = BumpAllocator<Node, Device>;

  TNL::Containers::Array<Node, Device> mMemory;

  Allocator mAlloc;
  Latch mLatch;
  DeviceVariable<Node *, Devices::Cuda> mRootVar;

  const size_t MaxBlockSize = 256;

public:
  using NodeType = Node;
  using DeviceType = Device;

  BTreeContainer() : mMemory(1), mAlloc(mMemory.getView()), mLatch(), mRootVar(nullptr) {
    static_assert(std::is_trivially_copyable<Key>::value, "Key must be trivially copyable");
    static_assert(std::is_trivially_copyable<Value>::value, "Value must be trivially copyable");
  };

  BTreeContainer(size_t memorySize) : BTreeContainer() { setSize(memorySize); };

  ~BTreeContainer() { mMemory.reset(); }

  void init() {
    BTreeInitKernel<Key, Value, Order, BTree><<<1, 1>>>(mRootVar, mAlloc);
    cudaDeviceSynchronize();
  }

  void setSize(size_t items) {
    Node *oldData = mMemory.getData();
    mMemory.resize(getContainerSize(items, Order));
    mAlloc.setView(mMemory.getView());

    if (*mRootVar != nullptr) {
      mRootVar = mMemory.getData() + (oldData - *mRootVar);
    }
  }

  void init(TNL::Containers::ArrayView<Key, Device> keys,
            TNL::Containers::ArrayView<Value, Device> values) {
    TNL_ASSERT_EQ(keys.getSize(), values.getSize(), "Values and keys must be the same size");

    const size_t Offset = 1;
    const size_t BulkNodeSize = Order - Offset;

    DeviceVariable<Node *, Devices::Cuda> rootBase;

    size_t expectNodeCount = (keys.getSize() + BulkNodeSize - 1) / BulkNodeSize;
    BTreeBulkAllocate<<<1, 1>>>(rootBase, expectNodeCount, mAlloc);
    cudaDeviceSynchronize();

    Containers::Array<Key, Devices::Cuda> resultKeys(expectNodeCount);
    Containers::Array<Node *, Devices::Cuda> resultChildren(expectNodeCount);

    size_t numBlocks = (keys.getSize() + MaxBlockSize - 1) / MaxBlockSize;
    size_t blockSize = MaxBlockSize;

    BTreeBulkCreateLeafKernel<<<numBlocks, blockSize>>>(rootBase, keys.getConstView(),
                                                        values.getView(), resultKeys.getView(),
                                                        resultChildren.getView(), mAlloc);

    cudaDeviceSynchronize();

    while (expectNodeCount > 1) {
      Containers::Array<Key, Devices::Cuda> inputKeys;
      Containers::Array<Node *, Devices::Cuda> inputChildren;

      inputKeys = resultKeys.getConstView(1, expectNodeCount);
      inputChildren = resultChildren.getConstView(0, expectNodeCount);

      expectNodeCount = (expectNodeCount + BulkNodeSize - 1) / BulkNodeSize;

      BTreeBulkAllocate<<<1, 1>>>(rootBase, expectNodeCount, mAlloc);
      cudaDeviceSynchronize();

      BTreeBulkCreateInternalKernel<<<numBlocks, blockSize>>>(
          rootBase, inputKeys.getConstView(), inputChildren.getView(), resultKeys.getView(),
          resultChildren.getView(), mAlloc);
      cudaDeviceSynchronize();
    }

    auto resultView = resultChildren.getView();
    auto &rootVar = mRootVar;
    auto assignRoot = [resultView, rootVar] __cuda_callable__(int i) mutable {
      rootVar = resultView.getElement(i);
    };

    Algorithms::ParallelFor<Devices::Cuda>::exec(0, 1, assignRoot);
  }

  void insert(TNL::Containers::ArrayView<Key, Device> keys,
              TNL::Containers::ArrayView<Value, Device> values) {
    TNL_ASSERT_EQ(keys.getSize(), values.getSize(), "Values and keys must be the same size");

    size_t keySize = keys.getSize();
    size_t numBlocks = (keySize + MaxBlockSize - 1) / MaxBlockSize;
    size_t blockSize = MaxBlockSize;

    BTreeInsertKernel<Key, Value, Order, BTree>
        <<<numBlocks, blockSize>>>(mRootVar, keys.getConstView(), values.getView(), mAlloc, mLatch);

    cudaDeviceSynchronize();
  }

  void remove(TNL::Containers::ArrayView<Key, Device> keys) {
    size_t keySize = keys.getSize();
    size_t numBlocks = (keySize + MaxBlockSize - 1) / MaxBlockSize;
    size_t blockSize = MaxBlockSize;

    BTreeRemoveKernel<Key, Value, Order, BTree>
        <<<numBlocks, blockSize>>>(mRootVar, keys.getConstView(), mLatch);

    cudaDeviceSynchronize();
  }

  void find(TNL::Containers::ArrayView<Key, Device> keys,
            TNL::Containers::ArrayView<Value, Device> result,
            TNL::Containers::ArrayView<bool, Device> mask) {
    size_t keySize = keys.getSize();
    size_t numBlocks = (keySize + MaxBlockSize - 1) / MaxBlockSize;
    size_t blockSize = MaxBlockSize;

    BTreeFindKernel<Key, Value, Order, BTree>
        <<<numBlocks, blockSize>>>(mRootVar, keys.getConstView(), result.getView(), mask.getView());

    cudaDeviceSynchronize();
  }
};