#pragma once
#include <TNL/Assert.h>
#include <TNL/Containers/Array.h>
#include <TNL/Cuda/CudaCallable.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/Devices/Host.h>
#include <TNL/Exceptions/CudaSupportMissing.h>
#include <TNL/Exceptions/NotImplementedError.h>
#include <type_traits>

#include "../BNodeLatch/Default.hpp"
#include "../Utils/Algorithms.hpp"
#include "../Utils/BumpAllocator.hpp"

using namespace TNL;

template <typename Key, typename Value, size_t Order, typename Device,
          template <typename _Key, typename _Value, size_t _Order, typename _Device> class BTree>
class BTreeContainer;

template <typename Key, typename Value, size_t Order,
          template <typename _Key, typename _Value, size_t _Order, typename _Device> class BTree>
class BTreeContainer<Key, Value, Order, Devices::Host, BTree> {
  using Device = Devices::Host;

  using Tree = BTree<Key, Value, Order, Device>;
  using Node = typename Tree::NodeType;
  using Latch = BNodeLatch<Node, Device>;
  using Allocator = BumpAllocator<Node, Device>;

  TNL::Containers::Array<Node, Device> mMemory;
  Allocator mAlloc;
  Latch mLatch;
  Node *mRoot;

public:
  using NodeType = Node;
  using DeviceType = Device;

  BTreeContainer() : mMemory(1), mAlloc(mMemory.getView()), mLatch(), mRoot(nullptr) {
    static_assert(std::is_trivially_copyable<Key>::value, "Key must be trivially copyable");
    static_assert(std::is_trivially_copyable<Value>::value, "Value must be trivially copyable");
  };

  BTreeContainer(size_t memorySize) : BTreeContainer() { setSize(memorySize); };

  ~BTreeContainer(){};

  void init() { mRoot = Tree::init(mAlloc); }

  void setSize(size_t items) {
    Node *oldData = mMemory.getData();
    mMemory.resize(getContainerSize(items, Order));
    mAlloc.setView(mMemory.getView());

    if (mRoot != nullptr) {
      Node *newRoot = mMemory.getData() + (oldData - mRoot);
      printf("Address: 0x%lx <-> 0x%lx\n", ADDR(mRoot), ADDR(newRoot));
      mRoot = newRoot;
    }
  }

  void insert(TNL::Containers::Array<Key, Device> &keys,
              TNL::Containers::Array<Value, Device> &values) {
    TNL_ASSERT_EQ(keys.getSize(), values.getSize(), "Values and keys must be the same size");

    for (size_t i = 0; i < keys.getSize(); ++i) {
      Tree::insert(mRoot, keys.getElement(i), values.getElement(i), mAlloc, mLatch);
    }
  }

  void remove(TNL::Containers::Array<Key, Device> &keys) {
    for (size_t i = 0; i < keys.getSize(); ++i) {
      Tree::remove(mRoot, keys.getElement(i), mLatch);
    }
  }

  void find(TNL::Containers::Array<Key, Device> &keys,
            TNL::Containers::Array<Value, Device> &result,
            TNL::Containers::Array<bool, Device> &mask) {

    auto keyView = keys.getConstView();
    auto maskView = mask.getView();

    result.forAllElements([=] __cuda_callable__(int i, Value &value) mutable {
      Value result;
      bool found = Tree::find(mRoot, keyView[i], result);
      maskView[i] = found;

      if (found) {
        value = result;
      }
    });
  }
};
