#pragma once
#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>
#include <type_traits>

#include <TNL/Atomic.h>
#include <TNL/Containers/Array.h>
#include <TNL/Containers/ArrayView.h>
#include <TNL/Cuda/CudaCallable.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/Devices/Host.h>
#include <TNL/Exceptions/CudaSupportMissing.h>
#include <TNL/TypeTraits.h>

#include "../Constants.hpp"

using namespace TNL;
using namespace TNL::Containers;

template <typename Object, typename Device> struct BumpAllocator;

template <typename Object> struct BumpAllocator<Object, Devices::Cuda> {
  int *mCount;
  int *mRefCnt;
  ArrayView<Object, Devices::Cuda> mData;

  BumpAllocator(ArrayView<Object, Devices::Cuda> view)
      : mCount(nullptr), mRefCnt(new int(1)), mData(view) {

#ifdef HAVE_CUDA
    cudaMalloc(&mCount, sizeof(int));
    cudaMemset(mCount, 0, sizeof(int));
#endif
  }

  void setView(ArrayView<Object, Devices::Cuda> view) { mData.bind(view); }

  BumpAllocator(const BumpAllocator &value)
      : mCount(value.mCount), mRefCnt(value.mRefCnt), mData(value.mData) {
    (*mRefCnt)++;
  }

  ~BumpAllocator() {
    (*mRefCnt)--;

#ifdef HAVE_CUDA
    if (*mRefCnt == 0) {
      cudaFree(mCount);
    }
#endif
  }

#ifdef __CUDA_ARCH__
  __device__ Object *allocate(int delta = 1) {
    int old = atomicAdd(mCount, delta);
    Object *inst = &(mData[old]);
    return inst;
  }
#else
  Object *allocate(int delta = 1) { throw TNL::Exceptions::CudaSupportMissing(); }
#endif // __CUDA_ARCH__
};

template <typename Object> struct BumpAllocator<Object, Devices::Host> {
  int mCount;
  ArrayView<Object, Devices::Host> mData;

  BumpAllocator(ArrayView<Object, Devices::Host> view) : mCount(0), mData(view) {}

  void setView(ArrayView<Object, Devices::Host> view) { mData.bind(view); }

  Object *allocate(int delta = 1) {
    Object &inst = mData[mCount];
    mCount += delta;
    return &inst;
  }
};