#pragma once
#include "../Constants.hpp"
#include <TNL/Cuda/CudaCallable.h>

#include <cstddef>
#include <cstdio>

__cuda_callable__ void inline printIndent(int level = 0) {
  for (int i = 0; i < level; ++i) {
    printf("  ");
  }
}

template <template <typename KeyType, typename ValueType, size_t Order> class BNode,
          typename KeyType, typename ValueType, size_t Order>
__cuda_callable__ void inline printNode(BNode<KeyType, ValueType, Order> *volatile node) {

#ifdef __CUDA_ARCH__
  printf("thr[%d,%d]: 0x%lx (leaf=%d, w=%d, link=0x%lx, len=%lu/%lu, "
         "highKey=%lu)",
         blockIdx.x, threadIdx.x, ADDR(node), node->mLeaf, node->mWriteLock, ADDR(node->mSibling),
         (unsigned long)node->mSize, (unsigned long)node->childSize(), node->mHighKey);
#else
  printf("0x%lx (leaf=%d, w=%d, link=0x%lx, len=%lu/%lu, "
         "highKey=%lu)",
         ADDR(node), node->mLeaf, node->mWriteLock, ADDR(node->mSibling),
         (unsigned long)node->mSize, (unsigned long)node->childSize(), node->mHighKey);
#endif
}

template <template <typename KeyType, typename ValueType, size_t Order> class BNode,
          typename KeyType, typename ValueType, size_t Order>
__cuda_callable__ void inline printNodeKeys(BNode<KeyType, ValueType, Order> *volatile node,
                                            int level = 0) {
  size_t keySize = node->mSize;
  printf("Keys: ");

  bool first = true;
  for (int i = 0; i < keySize; ++i) {
    KeyType key = node->mKeys[i];
    if (!first)
      printf(", ");
    printf("[%d]: %lu", i, key);
    first = false;
  }
}

template <template <typename KeyType, typename ValueType, size_t Order> class BNode,
          typename KeyType, typename ValueType, size_t Order>
__cuda_callable__ void inline printNodeValues(BNode<KeyType, ValueType, Order> *volatile node,
                                              int level = 0) {
  size_t keySize = node->mSize;
  printf("Values: ");

  bool first = true;
  for (int i = 0; i < keySize; ++i) {
    KeyType key = node->mValues[i];
    if (!first)
      printf(", ");
    printf("%lu", key);
    first = false;
  }
}

template <template <typename KeyType, typename ValueType, size_t Order> class BNode,
          typename KeyType, typename ValueType, size_t Order>
__cuda_callable__ void inline printTree(BNode<KeyType, ValueType, Order> *volatile node,
                                        int level = 0, int index = 0) {
  printIndent(level), printf("[%d] ", index), printNode(node), printf("\n");
  printIndent(level), printNodeKeys(node, level), printf("\n");
  if (node->mLeaf) {
    printIndent(level), printNodeValues(node, level), printf("\n");
  }

  size_t childSize = node->childSize();
  for (int i = 0; i < childSize; ++i) {
    BNode<KeyType, ValueType, Order> *volatile child = node->mChildren[i];
    if (child != nullptr) {
      printTree(child, level + 1, i);
    }
  }

  if (level == 0) {
    printf("\n");
  }
}

template <template <typename KeyType, typename ValueType, size_t Order> class BNode,
          typename KeyType, typename ValueType, size_t Order>
__cuda_callable__ void inline printTreeLeftmost(
    BNode<KeyType, ValueType, Order> *volatile node, int level = 0,
    BNode<KeyType, ValueType, Order> *volatile siblingParent = nullptr) {

  BNode<KeyType, ValueType, Order> *volatile current = node;
  while (current != nullptr) {
    printIndent(level), printNode(current), printf("\n");
    printIndent(level), printNodeKeys(current, level), printf("\n");

    if (current->mLeaf) {
      printIndent(level), printNodeValues(current, level), printf("\n");
    } else {
      printTreeLeftmost(current->mChildren[0], level + 1, current);
    }

    current = current->mSibling;
  }
}

__cuda_callable__ inline void printBits(unsigned int num) {
  unsigned int size = sizeof(unsigned int);
  unsigned int maxPow = 1 << (size * 8 - 1);
  int i = 0;
  for (; i < size; ++i) {
    for (; i < size * 8; ++i) {
      // print last bit and shift left.
      printf("%u, ", num & maxPow ? 1 : 0);
      num = num << 1;
    }
  }
}