#include "BLinkTree/Default.hpp"
#include "BPlusTree/Default.hpp"
#include "Utils/BumpAllocator.hpp"
#include "Utils/Print.hpp"

#include <TNL/Devices/Host.h>
#include <gtest/gtest.h>

#include <algorithm>
#include <bits/stdint-uintn.h>
#include <climits>
#include <list>
#include <random>

#include <deque>
#include <map>
#include <numeric>
#include <utility>

using KeyType = uint64_t;
using ValueType = uint64_t;
const size_t Order = 15;
const size_t KeyInf = UINT64_MAX;

template <template <typename _KeyType, typename _ValueType, size_t _Order, typename _Device>
          class Tree>
using TreeTemplate = Tree<KeyType, ValueType, Order, Devices::Host>;

template <typename C> class HostBTreeTest : public testing::Test {
public:
  using BTree = C;
  using Node = typename C::NodeType;
  using Operations = typename C::OperationsType;
  using Allocator = BumpAllocator<Node, Devices::Host>;
  using Latch = BNodeLatch<Node, Devices::Host>;

  void validateBTreeStructure(Node *root) {
    int maxLevel = 1;
    std::map<int, std::vector<Node *>> levels;
    std::deque<std::pair<int, Node *>> queue;

    queue.push_back(std::make_pair(1, root));

    while (queue.empty() == false) {
      auto current = queue.front();
      int currentLevel = current.first;
      Node *currentNode = current.second;
      queue.pop_front();

      // validate nodes have proper child size
      if (currentNode->mLeaf) {
        ASSERT_EQ(currentNode->childSize(), currentNode->mSize)
            << "Expecting children count match the item count on leaves";
      } else {
        ASSERT_EQ(currentNode->childSize(), currentNode->mSize + 1)
            << "Expecting children count to be +1 larger than item count";
      }

      maxLevel = std::max(maxLevel, currentLevel);
      levels[currentLevel].push_back(currentNode);

      if (!currentNode->mLeaf) {
        for (int i = 0; i < currentNode->childSize(); ++i) {
          queue.push_back(std::make_pair(currentLevel + 1, currentNode->mChildren[i]));
        }
      }
    }

    for (auto pair : levels) {
      size_t nodeCount = pair.second.size();

      // validate keys inside node are sorted
      Node *levelFirstKey = nullptr;
      KeyType levelPrev = KeyInf;

      for (auto node : pair.second) {
        if (levelFirstKey == nullptr && node->mSize > 0) {
          levelFirstKey = node;
          levelPrev = node->mKeys[0];
        }

        // validate inside node
        for (int i = 1; i < node->mSize; ++i) {
          KeyType prev = node->mKeys[i - 1];
          KeyType curr = node->mKeys[i];
          ASSERT_GE(curr, prev) << "Keys must be sorted";
          ASSERT_NE(curr, prev) << "Duplicate keys are not allowed";

          ASSERT_NE(prev, KeyInf) << "Must not be equal infinity";
          ASSERT_NE(curr, KeyInf) << "Must not be equal infinity";
        }

        // validate on entire level
        for (int i = 0; i < node->mSize; ++i) {
          if (node == levelFirstKey && i == 0) {
            continue;
          }

          KeyType curr = node->mKeys[i];
          ASSERT_NE(curr, KeyInf) << "Must not be equal infinity";
          ASSERT_GT(curr, levelPrev) << "Keys must be sorted on all levels";
          levelPrev = curr;
        }
      }

      // validate siblings
      for (int i = 1; i < nodeCount; ++i) {
        Node *prev = pair.second[i - 1];
        Node *curr = pair.second[i];

        ASSERT_EQ(prev->mSibling, curr) << "Nodes of same level must be side-linked together";

        if (i + 1 == nodeCount) {
          ASSERT_EQ(curr->mSibling, nullptr) << "Sibling of last item must point to nullptr";
        }
      }
    }

    // verify high key is properly set
    for (int level = maxLevel - 1; level > 0; --level) {
      size_t nodeCount = levels[level].size();
      for (int i = 0; i < nodeCount; ++i) {
        Node *node = levels[level][i];
        KeyType maxRealKey = 0;

        for (int keyIdx = 0; keyIdx < node->mSize; ++keyIdx) {
          KeyType key = node->mKeys[keyIdx];
          maxRealKey = std::max(key, maxRealKey);
        }

        for (int childIdx = 0; childIdx < node->childSize(); ++childIdx) {
          Node *child = node->mChildren[childIdx];

          for (int keyIdx = 0; keyIdx < child->mSize; ++keyIdx) {
            KeyType key = child->mKeys[keyIdx];
            maxRealKey = std::max(key, maxRealKey);
          }
        }

        if (i + 1 == nodeCount) {
          ASSERT_EQ(node->mHighKeyFlag, false) << "High key must be disabled";
        } else {
          ASSERT_EQ(node->mHighKeyFlag, true) << "High key must be enabled";
          ASSERT_LT(maxRealKey, node->mHighKey)
              << "High key must be greater than the maximum key of subtree";
        }
      }
    }
  }
};

using TreeTypes = ::testing::Types<TreeTemplate<BLinkTree>, TreeTemplate<BPlusTree>>;
TYPED_TEST_SUITE(HostBTreeTest, TreeTypes);

TYPED_TEST(HostBTreeTest, InsertSeqValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());
  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  std::vector<KeyType> keys(3000);
  std::iota(keys.begin(), keys.end(), 0);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  this->validateBTreeStructure(root);

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    ASSERT_TRUE(HostBTree::find(root, expectKey, result));
    ASSERT_EQ(expectKey, result) << "Expected key " << expectKey << " not equal to result "
                                 << result;
  }
}

TYPED_TEST(HostBTreeTest, InsertSeqRemoveValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());

  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  std::random_device rd;
  std::mt19937 g(rd());

  std::vector<KeyType> keys(3000);
  std::iota(keys.begin(), keys.end(), 0);
  std::shuffle(keys.begin(), keys.end(), g);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  std::vector<KeyType> removeList;

  for (int i = 0; i < keys.size(); ++i) {
    if (i % 3 == 0) {
      removeList.push_back(i);
    }
  }

  std::set<KeyType> removed(removeList.begin(), removeList.end());
  for (KeyType remove : removed) {
    HostBTree::remove(root, remove, latch);
  }

  this->validateBTreeStructure(root);

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    if (removed.count(expectKey)) {
      ASSERT_FALSE(HostBTree::find(root, expectKey, result));
    } else {
      ASSERT_TRUE(HostBTree::find(root, expectKey, result));
      ASSERT_EQ(expectKey, result)
          << "Expected key " << expectKey << " not equal to result " << result;
    }
  }
}

TYPED_TEST(HostBTreeTest, InsertRandomValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());

  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  std::vector<KeyType> keys(3000);
  std::iota(keys.begin(), keys.end(), 0);

  std::random_device rd;
  std::mt19937 g(rd());
  std::shuffle(keys.begin(), keys.end(), g);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  this->validateBTreeStructure(root);

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    ASSERT_TRUE(HostBTree::find(root, expectKey, result));
    ASSERT_EQ(expectKey, result) << "Expected key " << expectKey << " not equal to result "
                                 << result;
  }
}

TYPED_TEST(HostBTreeTest, InsertRandomRemoveValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());
  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  std::random_device rd;
  std::mt19937 g(rd());

  std::vector<KeyType> keys(3000);
  std::iota(keys.begin(), keys.end(), 0);
  std::shuffle(keys.begin(), keys.end(), g);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  std::vector<KeyType> removeList(keys);
  std::shuffle(removeList.begin(), removeList.end(), g);
  removeList.resize(keys.size() * 0.3);

  std::set<KeyType> removed(removeList.begin(), removeList.end());

  for (KeyType remove : removed) {
    HostBTree::remove(root, remove, latch);
  }

  this->validateBTreeStructure(root);

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    if (removed.count(expectKey)) {
      ASSERT_FALSE(HostBTree::find(root, expectKey, result));
    } else {
      ASSERT_TRUE(HostBTree::find(root, expectKey, result));
      ASSERT_EQ(expectKey, result)
          << "Expected key " << expectKey << " not equal to result " << result;
    }
  }
}

TYPED_TEST(HostBTreeTest, InsertCudaWrongValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());
  HostLatch latch;

  std::vector<KeyType> keys = {
      0,  32, 1,  2,  3,  33, 4,  34, 5,  35, 6,  36, 7,  37, 8,  38, 9,  10, 11, 12,
      13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 39, 28, 29, 30, 31,
  };

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  printTree(root);

  this->validateBTreeStructure(root);

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    ASSERT_TRUE(HostBTree::find(root, expectKey, result));
    ASSERT_EQ(expectKey, result) << "Expected key " << expectKey << " not equal to result "
                                 << result;
  }
}

TYPED_TEST(HostBTreeTest, InsertSidelinkValidation) {
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostOperations = typename TestFixture::Operations;
  using HostLatch = typename TestFixture::Latch;
  using HostBTree = typename TestFixture::BTree;

  TNL::Containers::Array<HostNode, Devices::Host> memory(100000);
  HostAlloc alloc(memory.getView());
  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  std::vector<KeyType> keys(3000);
  std::iota(keys.begin(), keys.end(), 0);

  std::random_device rd;
  std::mt19937 g(rd());
  std::shuffle(keys.begin(), keys.end(), g);

  for (KeyType key : keys) {
    HostBTree::insert(root, key, key, alloc, latch);
  }

  // validate proper result
  for (KeyType expectKey : keys) {
    ValueType result;

    ASSERT_TRUE(HostBTree::find(root, expectKey, result));
    ASSERT_EQ(expectKey, result) << "Expected key " << expectKey << " not equal to result "
                                 << result;
  }
}

int main(int argc, char **argv) {
  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}