#include "BLinkTree/Default.hpp"
#include "BPlusTree/Default.hpp"
#include "BTreeContainer/Default.hpp"
#include "Utils/Print.hpp"

#include <TNL/Assert.h>
#include <TNL/Exceptions/NotImplementedError.h>

#include <algorithm>
#include <bits/stdint-uintn.h>
#include <climits>
#include <cstdlib>
#include <iostream>
#include <list>
#include <numeric>
#include <random>
#include <sstream>
#include <type_traits>

#include <deque>
#include <map>
#include <numeric>
#include <utility>

#include <TNL/Containers/Array.h>
#include <TNL/Containers/ArrayView.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/Pointers/DevicePointer.h>

#include <cooperative_groups.h>
#include <cuda_profiler_api.h>

#include <gtest/gtest.h>

using namespace TNL;
using namespace TNL::Containers;

using KeyType = uint32_t;
using ValueType = uint32_t;
const size_t Order = 15;

template <template <typename _KeyType, typename _ValueType, size_t _Order, typename _Device>
          class Tree,
          typename Device>
using Uint32Types = std::pair<BTreeContainer<KeyType, ValueType, Order, Device, Tree>, Device>;

template <typename C> class ContainerTest : public testing::Test {
public:
  using Container = typename C::first_type;
  using Device = typename C::second_type;
};

using TreeTypes =
    ::testing::Types<Uint32Types<BPlusTree, Devices::Host>, Uint32Types<BPlusTree, Devices::Cuda>,
                     Uint32Types<BLinkTree, Devices::Host>, Uint32Types<BLinkTree, Devices::Cuda>>;
TYPED_TEST_SUITE(ContainerTest, TreeTypes);

std::random_device rd;
std::mt19937 g(rd());

TYPED_TEST(ContainerTest, InsertSeq) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> input(100);
  std::iota(input.begin(), input.end(), 0);

  Containers::Array<KeyType, Device> keys(input);
  Containers::Array<ValueType, Device> values(input);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);

  ASSERT_EQ(values, result);
  ASSERT_FALSE(mask.containsValue(false));
}

TYPED_TEST(ContainerTest, InsertRandom) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> input(1000);
  std::iota(input.begin(), input.end(), 0);
  std::shuffle(input.begin(), input.end(), g);

  Containers::Array<KeyType, Device> keys(input);
  Containers::Array<ValueType, Device> values(input);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);

  ASSERT_EQ(values, result);
  ASSERT_FALSE(mask.containsValue(false));
}

TYPED_TEST(ContainerTest, InsertRandomLarge) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> input(10000000);
  std::iota(input.begin(), input.end(), 0);
  std::shuffle(input.begin(), input.end(), g);

  Containers::Array<KeyType, Device> keys(input);
  Containers::Array<ValueType, Device> values(input);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);

  ASSERT_EQ(values, result);
  ASSERT_FALSE(mask.containsValue(false));
}

TYPED_TEST(ContainerTest, InsertSeqRemove) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> inputInsert(3000);
  std::vector<KeyType> inputRemove;

  std::vector<uint8_t> valuesRemove;
  std::iota(inputInsert.begin(), inputInsert.end(), 0);

  for (int i = 0; i < inputInsert.size(); ++i) {
    bool toRemove = i % 3 == 0;
    valuesRemove.push_back(!toRemove);

    if (toRemove) {
      inputRemove.push_back(i);
    }
  }

  Containers::Array<KeyType, Device> keys(inputInsert);
  Containers::Array<ValueType, Device> values(inputInsert);
  Containers::Array<KeyType, Device> remove(inputRemove);

  Containers::Array<bool, Device> expectMask(valuesRemove);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);
  expectMask = 1;

  ASSERT_EQ(result, values);
  ASSERT_EQ(mask, expectMask);

  container.remove(remove);
  container.find(keys, result, mask);
  expectMask = valuesRemove;

  ASSERT_EQ(mask, expectMask);
}

TYPED_TEST(ContainerTest, InsertRandomRemove) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> inputInsert(3000);
  std::iota(inputInsert.begin(), inputInsert.end(), 0);
  std::shuffle(inputInsert.begin(), inputInsert.end(), g);

  std::vector<KeyType> inputRemove(inputInsert);
  std::shuffle(inputRemove.begin(), inputRemove.end(), g);
  inputRemove.resize(inputRemove.size() * 0.3);

  std::set<KeyType> removeSet(inputRemove.begin(), inputRemove.end());
  std::vector<KeyType> removeMask;
  for (ValueType value : inputInsert) {
    removeMask.push_back(!removeSet.count(value));
  }

  Containers::Array<KeyType, Device> keys(inputInsert);
  Containers::Array<ValueType, Device> values(inputInsert);
  Containers::Array<KeyType, Device> remove(inputRemove);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> expectMask(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);
  expectMask = 1;

  ASSERT_EQ(mask, expectMask);
  ASSERT_EQ(result, values);

  container.remove(remove);
  container.find(keys, result, mask);
  expectMask = removeMask;

  ASSERT_EQ(mask, expectMask);
}

TYPED_TEST(ContainerTest, InsertRandomRemoveLarge) {
  using Device = typename TestFixture::Device;
  using Container = typename TestFixture::Container;

  Container container;

  std::vector<KeyType> inputInsert(3000000);
  std::iota(inputInsert.begin(), inputInsert.end(), 0);
  std::shuffle(inputInsert.begin(), inputInsert.end(), g);

  std::vector<KeyType> inputRemove(inputInsert);
  std::shuffle(inputRemove.begin(), inputRemove.end(), g);
  inputRemove.resize(inputRemove.size() * 0.3);

  std::set<KeyType> removeSet(inputRemove.begin(), inputRemove.end());
  std::vector<KeyType> removeMask;
  for (ValueType value : inputInsert) {
    removeMask.push_back(!removeSet.count(value));
  }

  Containers::Array<KeyType, Device> keys(inputInsert);
  Containers::Array<ValueType, Device> values(inputInsert);
  Containers::Array<KeyType, Device> remove(inputRemove);

  Containers::Array<ValueType, Device> result(keys.getSize());
  Containers::Array<bool, Device> expectMask(keys.getSize());
  Containers::Array<bool, Device> mask(keys.getSize());

  container.setSize(keys.getSize());
  container.init();

  container.insert(keys, values);
  container.find(keys, result, mask);
  expectMask = 1;

  ASSERT_EQ(mask, expectMask);
  ASSERT_EQ(result, values);

  container.remove(remove);
  container.find(keys, result, mask);
  expectMask = removeMask;

  ASSERT_EQ(mask, expectMask);
}

int main(int argc, char **argv) {
  cudaDeviceSetLimit(cudaLimitPrintfFifoSize, 1024 * 1024 * 1024);
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}