#include "BLinkTree/Default.hpp"
#include "BPlusTree/Default.hpp"
#include "Utils/BumpAllocator.hpp"
#include "Utils/Print.hpp"

#include <TNL/Devices/Host.h>
#include <gtest/gtest.h>

#include <algorithm>
#include <bits/stdint-uintn.h>

#include <climits>
#include <deque>
#include <list>
#include <map>
#include <numeric>
#include <random>
#include <utility>

using KeyType = uint64_t;
using ValueType = uint64_t;
const size_t KeyInf = UINT64_MAX;

template <typename BTree> class BTreeTestUtils {
  using Node = typename BTree::NodeType;
  using Operations = typename BTree::OperationsType;
  using Allocator = BumpAllocator<Node, Devices::Host>;
  using Latch = BNodeLatch<Node, Devices::Host>;

public:
  static Node *createNode(Allocator &alloc, bool leaf, Node *sibling, KeyType high, bool highFlag,
                          std::vector<KeyType> keys) {
    Node *node = alloc.allocate();
    Operations::init(node, leaf, sibling, highFlag);
    node->mHighKey = high;

    for (KeyType key : keys) {
      if (node->mLeaf) {
        Operations::appendKey(node, key, key);
      } else {
        Operations::appendKey(node, key);
      }
    }
    return node;
  }

  static std::deque<Node *> createNodeLevel(Allocator &alloc, Node *parent, bool leaf,
                                            std::vector<std::vector<KeyType>> nodeKeys,
                                            std::deque<Node *> prevLevel = {}) {
    std::deque<Node *> res;

    Node *prevNode = nullptr;
    KeyType high;
    bool highFlag = false;

    for (int i = nodeKeys.size() - 1; i >= 0; --i) {
      std::vector<KeyType> &keys = nodeKeys[i];
      Node *node = createNode(alloc, leaf, prevNode, high, highFlag, keys);

      prevNode = node;
      high = keys[0];
      highFlag = true;

      res.push_front(node);
    }

    int size = 0;
    for (auto child : res) {
      Operations::appendChild(parent, child, size);
      size++;
    }

    if (prevLevel.size() > 0) {
      Node *lastPrev = prevLevel.back();
      Node *firstCurr = res.front();
      lastPrev->mSibling = firstCurr;
      lastPrev->mHighKey = firstCurr->mKeys[0];
    }

    return res;
  }

  static std::vector<KeyType> iota(KeyType from, KeyType to) {
    std::vector<KeyType> res;
    for (KeyType i = from; i <= to; ++i) {
      res.push_back(i);
    }
    return res;
  }

  static std::vector<std::vector<KeyType>> iotaMap(std::vector<std::pair<KeyType, KeyType>> pairs) {
    std::vector<std::vector<KeyType>> res;
    for (auto pair : pairs) {
      res.push_back(iota(pair.first, pair.second));
    }
    return res;
  }

  static void assertNodeKeys(Node *node, std::vector<KeyType> expectKeys) {
    ASSERT_TRUE(node != nullptr);

    std::vector<KeyType> actualKeys;
    for (size_t i = 0; i < node->mSize; ++i) {
      actualKeys.push_back(node->mKeys[i]);
    }

    ASSERT_EQ(expectKeys, actualKeys);
  }

  static void assertNodeLevel(Node *parent, std::vector<std::vector<KeyType>> expectKeys) {
    ASSERT_TRUE(parent != nullptr);

    std::vector<std::vector<KeyType>> actualKeys;
    for (size_t cid = 0; cid < parent->childSize(); ++cid) {
      Node *child = parent->mChildren[cid];
      std::vector<KeyType> childKeys;
      for (size_t i = 0; i < child->mSize; ++i) {
        childKeys.push_back(child->mKeys[i]);
      }

      actualKeys.emplace_back(childKeys);
    }

    ASSERT_EQ(expectKeys, actualKeys);
  }
};

template <size_t Order,
          template <typename _KeyType, typename _ValueType, size_t _Order, typename _Device>
          class Tree>
using TreeTemplate = Tree<KeyType, ValueType, Order, Devices::Host>;

template <typename C> class BTreeTest : public testing::Test {
protected:
  void SetUp() override {
    stackMem.setSize(1000);
    memory.setSize(1000);
  }

  void TearDown() override {}

public:
  using BTree = C;
  using Node = typename C::NodeType;
  using Operations = typename C::OperationsType;
  using Allocator = BumpAllocator<Node, Devices::Host>;
  using Latch = BNodeLatch<Node, Devices::Host>;
  using TestUtils = BTreeTestUtils<BTree>;

  TNL::Containers::Array<typename C::NodeType, Devices::Host> memory;
  TNL::Containers::Array<typename C::NodeType *, Devices::Host> stackMem;
};

template <typename C> class LargeOrderTest : public BTreeTest<C> {};
using TreeTypes = ::testing::Types<TreeTemplate<15, BPlusTree>, TreeTemplate<15, BLinkTree>>;
TYPED_TEST_SUITE(LargeOrderTest, TreeTypes);

TYPED_TEST(LargeOrderTest, Empty) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  HostAlloc alloc(this->memory.getView());
  HostLatch latch;

  HostNode *root = alloc.allocate();
  HostOperations::init(root, true, nullptr, false);

  HostBTree::insert(root, 1, 1, alloc, latch);
  TestFixture::TestUtils::assertNodeKeys(root, {1});
}

TYPED_TEST(LargeOrderTest, FullRoot) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  HostAlloc alloc(this->memory.getView());
  HostLatch latch;

  HostNode *root = TestFixture::TestUtils::createNode(
      alloc, true, nullptr, KeyInf, false, {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14});

  ASSERT_TRUE(HostBTree::insert(root, 15, 15, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {7});
  TestFixture::TestUtils::assertNodeLevel(
      root, {{0, 1, 2, 3, 4, 5, 6}, {7, 8, 9, 10, 11, 12, 13, 14, 15}});

  ASSERT_TRUE(root->mChildren[0]->mLeaf == true);
  ASSERT_TRUE(root->mChildren[1]->mLeaf == true);
  ASSERT_TRUE(root->mChildren[0]->mSibling == root->mChildren[1]);
}

TYPED_TEST(LargeOrderTest, InsertNested) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  HostAlloc alloc(this->memory.getView());
  HostLatch latch;

  HostNode *root = TestFixture::TestUtils::createNode(alloc, false, nullptr, KeyInf, false, {7});
  TestFixture::TestUtils::createNodeLevel(alloc, root, true,
                                          {{1, 2, 3, 4, 5, 6}, {7, 8, 9, 10, 11, 12, 13, 14}});

  ASSERT_TRUE(HostBTree::insert(root, 15, 15, alloc, latch));
  ASSERT_TRUE(HostBTree::insert(root, 0, 0, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {7});
  TestFixture::TestUtils::assertNodeLevel(
      root, {{0, 1, 2, 3, 4, 5, 6}, {7, 8, 9, 10, 11, 12, 13, 14, 15}});
}

TYPED_TEST(LargeOrderTest, InsertFullNonIncrease) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  HostAlloc alloc(this->memory.getView());
  HostLatch latch;

  HostNode *root = TestFixture::TestUtils::createNode(alloc, false, nullptr, KeyInf, false, {7});
  auto rootChildren = TestFixture::TestUtils::createNodeLevel(
      alloc, root, true,
      {{0, 1, 2, 3, 4, 5, 6}, {7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21}});

  ASSERT_TRUE(HostBTree::insert(root, 22, 22, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {7, 14});
  TestFixture::TestUtils::assertNodeLevel(
      root,
      {{0, 1, 2, 3, 4, 5, 6}, {7, 8, 9, 10, 11, 12, 13}, {14, 15, 16, 17, 18, 19, 20, 21, 22}});
}

TYPED_TEST(LargeOrderTest, InsertFullIncrease) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  HostAlloc alloc(this->memory.getView());
  HostLatch latch;

  HostNode *root =
      TestFixture::TestUtils::createNode(alloc, false, nullptr, KeyInf, false,
                                         {7, 14, 21, 28, 35, 42, 49, 56, 63, 70, 77, 84, 91, 98});
  TestFixture::TestUtils::createNodeLevel(alloc, root, true,
                                          TestFixture::TestUtils::iotaMap({
                                              {0, 6},
                                              {7, 13},
                                              {14, 20},
                                              {21, 27},
                                              {28, 34},
                                              {35, 41},
                                              {42, 48},
                                              {49, 55},
                                              {56, 62},
                                              {63, 69},
                                              {70, 76},
                                              {77, 83},
                                              {84, 90},
                                              {91, 97},
                                              {98, 112},
                                          }));

  ASSERT_TRUE(HostBTree::insert(root, 113, 113, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {56});
  TestFixture::TestUtils::assertNodeLevel(
      root, {{7, 14, 21, 28, 35, 42, 49}, {63, 70, 77, 84, 91, 98, 105}});

  TestFixture::TestUtils::assertNodeLevel(
      root->mChildren[0],
      TestFixture::TestUtils::iotaMap(
          {{0, 6}, {7, 13}, {14, 20}, {21, 27}, {28, 34}, {35, 41}, {42, 48}, {49, 55}}));

  TestFixture::TestUtils::assertNodeLevel(
      root->mChildren[1],
      TestFixture::TestUtils::iotaMap(
          {{56, 62}, {63, 69}, {70, 76}, {77, 83}, {84, 90}, {91, 97}, {98, 104}, {105, 113}}));
}

template <typename C> class SmallOrderTest : public BTreeTest<C> {};
using SmallTreeTypes = ::testing::Types<TreeTemplate<4, BPlusTree>, TreeTemplate<4, BLinkTree>>;
TYPED_TEST_SUITE(SmallOrderTest, SmallTreeTypes);

TYPED_TEST(SmallOrderTest, InsertSmallTwoLevelEnd) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  TNL::Containers::Array<HostNode, Devices::Host> memory(1000);
  TNL::Containers::Array<HostNode *, Devices::Host> stackMem(1000);

  HostAlloc alloc(memory.getView());
  HostLatch latch;

  HostNode *root = TestFixture::TestUtils::createNode(alloc, false, nullptr, KeyInf, false, {7});
  TestFixture::TestUtils::createNodeLevel(alloc, root, false, {{3, 5}, {9, 11, 13}});
  auto lt = TestFixture::TestUtils::createNodeLevel(alloc, root->mChildren[0], true,
                                                    {{1, 2}, {3, 4}, {5, 6}});
  TestFixture::TestUtils::createNodeLevel(alloc, root->mChildren[1], true,
                                          {{7, 8}, {9, 10}, {11, 12}, {13, 14, 15, 16}}, lt);

  ASSERT_TRUE(HostBTree::insert(root, 17, 17, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {7, 11});
  TestFixture::TestUtils::assertNodeLevel(root, {{3, 5}, {9}, {13, 15}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[0], {{1, 2}, {3, 4}, {5, 6}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[1], {{7, 8}, {9, 10}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[2], {{11, 12}, {13, 14}, {15, 16, 17}});
}

TYPED_TEST(SmallOrderTest, InsertSmallTwoLevelMid) {
  using HostBTree = typename TestFixture::BTree;
  using HostNode = typename TestFixture::Node;
  using HostAlloc = typename TestFixture::Allocator;
  using HostLatch = typename TestFixture::Latch;
  using HostOperations = typename TestFixture::Operations;

  TNL::Containers::Array<HostNode, Devices::Host> memory(1000);
  TNL::Containers::Array<HostNode *, Devices::Host> stackMem(1000);

  HostAlloc alloc(memory.getView());
  HostLatch latch;

  HostNode *root = TestFixture::TestUtils::createNode(alloc, false, nullptr, KeyInf, false, {7});
  TestFixture::TestUtils::createNodeLevel(alloc, root, false, {{3, 5}, {9, 11, 20}});
  auto lt = TestFixture::TestUtils::createNodeLevel(alloc, root->mChildren[0], true,
                                                    {{1, 2}, {3, 4}, {5, 6}});
  TestFixture::TestUtils::createNodeLevel(alloc, root->mChildren[1], true,
                                          {{7, 8}, {9, 10}, {11, 12}, {20, 21, 22}}, lt);

  ASSERT_TRUE(HostBTree::insert(root, 13, 13, alloc, latch));

  TestFixture::TestUtils::assertNodeKeys(root, {7, 11});
  TestFixture::TestUtils::assertNodeLevel(root, {{3, 5}, {9}, {20}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[0], {{1, 2}, {3, 4}, {5, 6}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[1], {{7, 8}, {9, 10}});
  TestFixture::TestUtils::assertNodeLevel(root->mChildren[2], {{11, 12, 13}, {20, 21, 22}});
}

int main(int argc, char **argv) {
  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}